"""Script to run MCMC cosmological sampling for individual lenses, using the BNN posterior

It borrows heavily from the `catalogue modelling.ipynb` notebook in Lenstronomy Extensions, which you can find `here <https://github.com/sibirrer/lenstronomy_extensions/blob/master/lenstronomy_extensions/Notebooks/catalogue%20modelling.ipynb>`_.

Example
-------
To run this script, pass in the path to the user-defined inference config file as the argument::
    
    $ python h0rton/infer_h0_mcmc_default.py mcmc_default.json

"""
import os
import time
from tqdm import tqdm
import gc
from ast import literal_eval
import numpy as np
import pandas as pd
import torch
from torch.utils.data import DataLoader
from lenstronomy.Workflow.fitting_sequence import FittingSequence
from lenstronomy.Cosmo.lcdm import LCDM
import baobab.sim_utils.metadata_utils as metadata_utils
from baobab import BaobabConfig
import h0rton.models
from h0rton.configs import TrainValConfig, TestConfig
import h0rton.losses
import h0rton.train_utils as train_utils
import h0rton.script_utils as script_utils
from h0rton.h0_inference import h0_utils, plotting_utils, mcmc_utils
from h0rton.trainval_data import XYData

def main():
    args = script_utils.parse_inference_args()
    test_cfg = TestConfig.from_file(args.test_config_file_path)
    baobab_cfg = BaobabConfig.from_file(test_cfg.data.test_baobab_cfg_path)
    cfg = TrainValConfig.from_file(test_cfg.train_val_config_file_path)
    # Set device and default data type
    device = torch.device(test_cfg.device_type)
    if device.type == 'cuda':
        torch.set_default_tensor_type('torch.cuda.' + cfg.data.float_type)
    else:
        torch.set_default_tensor_type('torch.' + cfg.data.float_type)
    script_utils.seed_everything(test_cfg.global_seed)
    
    ############
    # Data I/O #
    ############
    train_data = XYData(is_train=True, 
                        Y_cols=cfg.data.Y_cols, 
                        float_type=cfg.data.float_type, 
                        define_src_pos_wrt_lens=cfg.data.define_src_pos_wrt_lens, 
                        rescale_pixels=cfg.data.rescale_pixels, 
                        rescale_pixels_type=cfg.data.rescale_pixels_type,
                        log_pixels=cfg.data.log_pixels, 
                        add_pixel_noise=cfg.data.add_pixel_noise, 
                        eff_exposure_time=cfg.data.eff_exposure_time, 
                        train_Y_mean=None, 
                        train_Y_std=None, 
                        train_baobab_cfg_path=cfg.data.train_baobab_cfg_path, 
                        val_baobab_cfg_path=None, 
                        for_cosmology=False)
    # Define val data and loader
    test_data = XYData(is_train=False, 
                       Y_cols=cfg.data.Y_cols, 
                       float_type=cfg.data.float_type, 
                       define_src_pos_wrt_lens=cfg.data.define_src_pos_wrt_lens, 
                       rescale_pixels=cfg.data.rescale_pixels, 
                       rescale_pixels_type=cfg.data.rescale_pixels_type,
                       log_pixels=cfg.data.log_pixels, 
                       add_pixel_noise=cfg.data.add_pixel_noise, 
                       eff_exposure_time=cfg.data.eff_exposure_time, 
                       train_Y_mean=train_data.train_Y_mean, 
                       train_Y_std=train_data.train_Y_std, 
                       train_baobab_cfg_path=cfg.data.train_baobab_cfg_path, 
                       val_baobab_cfg_path=test_cfg.data.test_baobab_cfg_path, 
                       for_cosmology=True)
    master_truth = test_data.Y_df
    master_truth = metadata_utils.add_qphi_columns(master_truth)
    master_truth = metadata_utils.add_gamma_psi_ext_columns(master_truth)
    # Figure out how many lenses BNN will predict on (must be consecutive)
    if test_cfg.data.lens_indices is None:
        if args.lens_indices_path is None:
            # Test on all n_test lenses in the test set
            n_test = test_cfg.data.n_test 
            lens_range = range(n_test)
        else:
            # Test on the lens indices in a text file at the specified path
            lens_range = []
            with open(args.lens_indices_path, "r") as f:
                for line in f:
                    lens_range.append(int(line.strip()))
            n_test = len(lens_range)
            print("Performing H0 inference on {:d} specified lenses...".format(n_test))
    else:
        if args.lens_indices_path is None:
            # Test on the lens indices specified in the test config file
            lens_range = test_cfg.data.lens_indices
            n_test = len(lens_range)
            print("Performing H0 inference on {:d} specified lenses...".format(n_test))
        else:
            raise ValueError("Specific lens indices were specified in both the test config file and the command-line argument.")
    batch_size = max(lens_range) + 1
    test_loader = DataLoader(test_data, batch_size=batch_size, shuffle=False, drop_last=True)
    # Output directory into which the H0 histograms and H0 samples will be saved
    out_dir = test_cfg.out_dir
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
        print("Destination folder path: {:s}".format(out_dir))
    else:
        raise OSError("Destination folder already exists.")

    #####################
    # Parameter penalty #
    #####################
    # Instantiate original loss function with all BNN-predicted params
    orig_Y_cols = cfg.data.Y_cols
    loss_fn = getattr(h0rton.losses, cfg.model.likelihood_class)(Y_dim=test_data.Y_dim, 
                                                                 device=device)
    # Not all predicted params will be sampled via MCMC
    params_to_remove = [] #'lens_light_R_sersic', 'src_light_R_sersic'] 
    mcmc_Y_cols = [col for col in orig_Y_cols if col not in params_to_remove]
    mcmc_Y_dim = len(mcmc_Y_cols)
    # Instantiate loss function with just the MCMC params
    mcmc_loss_fn = getattr(h0rton.losses, cfg.model.likelihood_class)(Y_dim=test_data.Y_dim - len(params_to_remove), 
                                                                      device=device)
    remove_param_idx, remove_idx = mcmc_utils.get_idx_for_params(mcmc_loss_fn.out_dim, 
                                                                 orig_Y_cols, 
                                                                 params_to_remove, 
                                                                 cfg.model.likelihood_class)
    mcmc_train_Y_mean = np.delete(train_data.train_Y_mean, remove_param_idx)
    mcmc_train_Y_std = np.delete(train_data.train_Y_std, remove_param_idx)
    parameter_penalty = mcmc_utils.HybridBNNPenalty(mcmc_Y_cols, cfg.model.likelihood_class, mcmc_train_Y_mean, mcmc_train_Y_std, test_cfg.h0_posterior.exclude_velocity_dispersion, device)
    custom_logL_addition = parameter_penalty.evaluate
    null_spread = False

    ###################
    # BNN predictions #
    ###################
    # Instantiate BNN model
    net = getattr(h0rton.models, cfg.model.architecture)(num_classes=loss_fn.out_dim, dropout_rate=cfg.model.dropout_rate)
    net.to(device)
    # Load trained weights from saved state
    net, epoch = train_utils.load_state_dict_test(test_cfg.state_dict_path, net, cfg.optim.n_epochs, device)
    # When only generating BNN predictions (and not running MCMC), we can afford more n_dropout
    # otherwise, we fix n_dropout = mcmc_Y_dim + 1
    if test_cfg.export.pred:
        n_dropout = 20
        n_samples_per_dropout = test_cfg.numerics.mcmc.walkerRatio
    else:
        n_walkers = test_cfg.numerics.mcmc.walkerRatio*(mcmc_Y_dim + 1) # (BNN params + D_dt) times walker ratio
        n_dropout = n_walkers//test_cfg.numerics.mcmc.walkerRatio
        n_samples_per_dropout = test_cfg.numerics.mcmc.walkerRatio
    # Initialize arrays that will store samples and BNN predictions
    init_pos = np.empty([batch_size, n_dropout, n_samples_per_dropout, mcmc_Y_dim])
    mcmc_pred = np.empty([batch_size, n_dropout, mcmc_loss_fn.out_dim])
    with torch.no_grad():
        net.train()
        # Send some empty forward passes through the test data without backprop to adjust batchnorm weights
        # (This is often not necessary. Beware if using for just 1 lens.)
        for nograd_pass in range(5):
            for X_, Y_ in test_loader:
                X = X_.to(device)
                _ = net(X)
        # Obtain MC dropout samples
        for d in range(n_dropout):
            net.eval()
            for X_, Y_ in test_loader:
                X = X_.to(device)
                Y = Y_.to(device)
                pred = net(X)
                break
            mcmc_pred_d = pred.cpu().numpy()
            # Replace BNN posterior's primary gaussian mean with truth values
            if test_cfg.lens_posterior_type == 'default_with_truth_mean':
                mcmc_pred_d[:, :len(mcmc_Y_cols)] = Y[:, :len(mcmc_Y_cols)].cpu().numpy()
            # Leave only the MCMC parameters in pred
            mcmc_pred_d = mcmc_utils.remove_parameters_from_pred(mcmc_pred_d, remove_idx, return_as_tensor=False)
            # Populate pred that will define the MCMC penalty function
            mcmc_pred[:, d, :] = mcmc_pred_d
            # Instantiate posterior to generate BNN samples, which will serve as initial positions for walkers
            bnn_post = getattr(h0rton.h0_inference.gaussian_bnn_posterior_cpu, loss_fn.posterior_name + 'CPU')(mcmc_Y_dim,  mcmc_train_Y_mean, mcmc_train_Y_std)
            bnn_post.set_sliced_pred(mcmc_pred_d)
            init_pos[:, d, :, :] = bnn_post.sample(n_samples_per_dropout, sample_seed=test_cfg.global_seed+d) # contains just the lens model params, no D_dt
            gc.collect()
    # Terminate right after generating BNN predictions (no MCMC)
    if test_cfg.export.pred:
        import sys
        samples_path = os.path.join(out_dir, 'samples.npy')
        np.save(samples_path, init_pos)
        sys.exit()

    #############
    # MCMC loop #
    #############
    # Convolve MC dropout iterates with aleatoric samples
    init_pos = init_pos.transpose(0, 3, 1, 2).reshape([batch_size, mcmc_Y_dim, -1]).transpose(0, 2, 1) # [batch_size, n_samples, mcmc_Y_dim]
    init_D_dt = np.random.uniform(0.0, 15000.0, size=(batch_size, n_walkers, 1))
    pred_mean = np.mean(init_pos, axis=1) # [batch_size, mcmc_Y_dim]
    # Define assumed model profiles
    kwargs_model = dict(lens_model_list=['PEMD', 'SHEAR'],
                        point_source_model_list=['SOURCE_POSITION'],
                        source_light_model_list=['SERSIC_ELLIPSE'])
    astro_sig = test_cfg.image_position_likelihood.sigma # astrometric uncertainty
    # Get H0 samples for each system
    if not test_cfg.time_delay_likelihood.baobab_time_delays:
        if 'abcd_ordering_i' not in master_truth:
            raise ValueError("If the time delay measurements were not generated using Baobab, the user must specify the order of image positions in which the time delays are listed, in order of increasing dec.")
    kwargs_lens_eqn_solver = {'min_distance': 0.05, 'search_window': baobab_cfg.instrument['pixel_scale']*baobab_cfg.image['num_pix'], 'num_iter_max': 200}

    total_progress = tqdm(total=n_test)
    realized_time_delays = pd.read_csv(test_cfg.error_model.realized_time_delays, index_col=None)
    # For each lens system...
    for i, lens_i in enumerate(lens_range):
        # Each lens gets a unique random state for time delay measurement error realizations.
        #rs_lens = np.random.RandomState(lens_i) # replaced with externally rendered time delays
        ###########################
        # Relevant data and prior #
        ###########################
        data_i = master_truth.iloc[lens_i].copy()
        # Set BNN pred defining parameter penalty for this lens, batch processes across n_dropout
        parameter_penalty.set_bnn_post_params(mcmc_pred[lens_i, :, :])
        # Initialize lens model params walkers at the predictive mean
        init_info = dict(zip(mcmc_Y_cols, pred_mean[lens_i, :]*mcmc_train_Y_std + mcmc_train_Y_mean))
        lcdm = LCDM(z_lens=data_i['z_lens'], z_source=data_i['z_src'], flat=True)
        true_img_dec = literal_eval(data_i['y_image'])
        n_img = len(true_img_dec)
        measured_td_sig = test_cfg.time_delay_likelihood.sigma
        measured_td_wrt0 = np.array(literal_eval(realized_time_delays.iloc[lens_i]['measured_td_wrt0']))
        kwargs_data_joint = dict(
                                 time_delays_measured=measured_td_wrt0,
                                 time_delays_uncertainties=measured_td_sig,
                                 )

        #############################
        # Parameter init and bounds #
        #############################
        lens_kwargs = mcmc_utils.get_lens_kwargs(init_info, null_spread=null_spread)
        ps_kwargs = mcmc_utils.get_ps_kwargs_src_plane(init_info, astro_sig)
        src_light_kwargs = mcmc_utils.get_light_kwargs(init_info['src_light_R_sersic'], null_spread=null_spread)
        special_kwargs = mcmc_utils.get_special_kwargs(n_img, astro_sig) # image position offset and time delay distance, aka the "special" parameters
        kwargs_params = {'lens_model': lens_kwargs,
                         'point_source_model': ps_kwargs,
                         'source_model': src_light_kwargs,
                         'special': special_kwargs,}
        if test_cfg.numerics.solver_type == 'NONE':
            solver_type = 'NONE'
        else:
            solver_type = 'PROFILE_SHEAR' if n_img == 4 else 'CENTER'
        #solver_type = 'NONE'
        kwargs_constraints = {'num_point_source_list': [n_img],  
                              'Ddt_sampling': True,
                              'solver_type': solver_type,}

        kwargs_likelihood = {'time_delay_likelihood': True,
                             'sort_images_by_dec': True,
                             'prior_lens': [],
                             'prior_special': [],
                             'check_bounds': True, 
                             'check_matched_source_position': False,
                             'source_position_tolerance': 0.01,
                             'source_position_sigma': 0.01,
                             'source_position_likelihood': False,
                             'custom_logL_addition': custom_logL_addition,
                             'kwargs_lens_eqn_solver': kwargs_lens_eqn_solver}

        ###########################
        # MCMC posterior sampling #
        ###########################
        fitting_seq = FittingSequence(kwargs_data_joint, kwargs_model, kwargs_constraints, kwargs_likelihood, kwargs_params, verbose=False, mpi=False)
        if i == 0:
            param_class = fitting_seq._updateManager.param_class
            n_params, param_class_Y_cols = param_class.num_param()
            init_pos = mcmc_utils.reorder_to_param_class(mcmc_Y_cols, param_class_Y_cols, init_pos, init_D_dt)
        # MCMC sample from the post-processed BNN posterior jointly with cosmology
        lens_i_start_time = time.time()
        if test_cfg.lens_posterior_type == 'default':
            test_cfg.numerics.mcmc.update(init_samples=init_pos[lens_i, :, :])
        fitting_kwargs_list_mcmc = [['MCMC', test_cfg.numerics.mcmc]]
        #try:
        with script_utils.HiddenPrints():
            chain_list_mcmc = fitting_seq.fit_sequence(fitting_kwargs_list_mcmc)
            kwargs_result_mcmc = fitting_seq.best_fit()
        lens_i_end_time = time.time()
        inference_time = (lens_i_end_time - lens_i_start_time)/60.0 # min

        #############################
        # Plotting the MCMC samples #
        #############################
        # sampler_type : 'EMCEE'
        # samples_mcmc : np.array of shape `[n_mcmc_eval, n_params]`
        # param_mcmc : list of str of length n_params, the parameter names
        sampler_type, samples_mcmc, param_mcmc, _  = chain_list_mcmc[0]
        new_samples_mcmc = mcmc_utils.postprocess_mcmc_chain(kwargs_result_mcmc, samples_mcmc, kwargs_model, lens_kwargs[2], ps_kwargs[2], src_light_kwargs[2], special_kwargs[2], kwargs_constraints)
        # Plot D_dt histogram
        D_dt_samples = new_samples_mcmc['D_dt'].values
        true_D_dt = lcdm.D_dt(H_0=data_i['H0'], Om0=0.3)
        data_i['D_dt'] = true_D_dt
        # Export D_dt samples for this lens
        lens_inference_dict = dict(
                                   D_dt_samples=D_dt_samples, # kappa_ext=0 for these samples
                                   inference_time=inference_time,
                                   true_D_dt=true_D_dt, 
                                   )
        lens_inference_dict_save_path = os.path.join(out_dir, 'D_dt_dict_{0:04d}.npy'.format(lens_i))
        np.save(lens_inference_dict_save_path, lens_inference_dict)
        # Optionally export the MCMC samples
        if test_cfg.export.mcmc_samples:
            mcmc_samples_path = os.path.join(out_dir, 'mcmc_samples_{0:04d}.csv'.format(lens_i))
            new_samples_mcmc.to_csv(mcmc_samples_path, index=None)
        # Optionally export the D_dt histogram
        if test_cfg.export.D_dt_histogram:
            cleaned_D_dt_samples = h0_utils.remove_outliers_from_lognormal(D_dt_samples, 3)
            _ = plotting_utils.plot_D_dt_histogram(cleaned_D_dt_samples, lens_i, true_D_dt, save_dir=out_dir)
        # Optionally export the plot of MCMC chain
        if test_cfg.export.mcmc_chain:
            mcmc_chain_path = os.path.join(out_dir, 'mcmc_chain_{0:04d}.png'.format(lens_i))
            plotting_utils.plot_mcmc_chain(chain_list_mcmc, mcmc_chain_path)
        # Optionally export posterior cornerplot of select lens model parameters with D_dt
        if test_cfg.export.mcmc_corner:
            mcmc_corner_path = os.path.join(out_dir, 'mcmc_corner_{0:04d}.png'.format(lens_i))
            plotting_utils.plot_mcmc_corner(new_samples_mcmc[test_cfg.export.mcmc_cols], data_i[test_cfg.export.mcmc_cols], test_cfg.export.mcmc_col_labels, mcmc_corner_path)
        total_progress.update(1)
        gc.collect()
    realized_time_delays.to_csv(os.path.join(out_dir, 'realized_time_delays.csv'), index=None)
    total_progress.close()

if __name__ == '__main__':
    #import cProfile
    #pr = cProfile.Profile()
    #pr.enable()
    main()
    #pr.disable()
    #pr.print_stats(sort='cumtime')


import numpy as np
__all__ = ['reorder_to_tdlmc']

def reorder_to_tdlmc(abcd_ordering_i, ra_img, dec_img, time_delays):
    """Reorder the list of ra, dec, and time delays to conform to the
    order in the TDLMC challenge

    Parameters
    ----------
    abcd_ordering_i : array-like
        ABCD in an increasing dec order if the keys ABCD mapped to values 0123, respectively, e.g. [3, 1, 0, 2] if D (value 3) is lowest, B (value 1) is second lowest
    ra_img : array-like
        list of ra from lenstronomy
    dec_img : array-like
        list of dec from lenstronomy, in the order specified by `ra_img`
    time_delays : array-like
        list of time delays from lenstronomy, in the order specified by `ra_img`

    Returns
    -------
    tuple
        tuple of (reordered ra, reordered_dec, reordered time delays)

    """
    ra_img = np.array(ra_img)
    dec_img = np.array(dec_img)
    time_delays = np.array(time_delays)
    # Order ra_pos, dec_pos, time_delays in increasing dec order 
    increasing_dec_i = np.argsort(dec_img)
    ra_img = ra_img[increasing_dec_i]
    dec_img = dec_img[increasing_dec_i]
    time_delays = time_delays[increasing_dec_i]
    # Reorder to get it in ABCD
    ra_img = ra_img[abcd_ordering_i]
    dec_img = dec_img[abcd_ordering_i]
    time_delays = time_delays[abcd_ordering_i]

    return (ra_img, dec_img, time_delays)





# -*- coding: utf-8 -*-
"""Generating a cosmological summary of the H0 or D_dt H0_samples

Example
-------
To run this script, pass in the version ID and the sampling method as the argument::
    
    $ python summarize.py 21 simple_mc_default

The summary will be saved to the same directory level as the sample directory.

"""
import os
import numpy as np
import pandas as pd
import argparse
import scipy.stats
from baobab.configs import BaobabConfig
from h0rton.configs import TestConfig
import h0rton.h0_inference.h0_utils as h0_utils
import h0rton.tdlmc_utils as tdlmc_utils

def parse_args():
    """Parse command-line arguments
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('version_id', help='version ID', type=int)
    parser.add_argument('sampling_method', help='the sampling method (one of simple_mc_default, mcmc_default, hybrid', type=str)
    parser.add_argument('--rung_idx', help='the TDLMC rung index, if H0rton was run on TDLMC data', type=int, default=None)
    args = parser.parse_args()
    return args

def main():
    args = parse_args()
    # Folder where all the H0 samples live
    samples_dir = '/home/jwp/stage/sl/h0rton/experiments/v{:d}/{:s}'.format(args.version_id, args.sampling_method)
    # Read in test cfg for this version and sampling method
    test_cfg_path = os.path.join(samples_dir, '..', '{:s}.json'.format(args.sampling_method))
    test_cfg = TestConfig.from_file(test_cfg_path)
    if 'mcmc_default' in args.sampling_method:
        summarize_mcmc(samples_dir, test_cfg, 'mcmc_default', args.rung_idx)
    elif args.sampling_method == 'hybrid':
        summarize_mcmc(samples_dir, test_cfg, 'hybrid')
    elif args.sampling_method == 'simple_mc_default':
        summarize_simple_mc_default(samples_dir, test_cfg)
    else:
        raise ValueError("This sampling method is not supported. Choose one of [simple_mc_default, mcmc_default, hybrid].")

def summarize_simple_mc_default(samples_dir, test_cfg):
    """Summarize the output of simple_mc_default, i.e. the uniform H0 samples with corresponding weights

    """
    H0_dicts = [f for f in os.listdir(samples_dir) if f.startswith('h0_dict')]
    H0_dicts.sort()
    # Read in the redshift columns of metadata
    baobab_cfg = BaobabConfig.from_file(test_cfg.data.test_baobab_cfg_path)
    metadata_path = os.path.join(baobab_cfg.out_dir, 'metadata.csv')
    meta = pd.read_csv(metadata_path, index_col=None, usecols=['z_lens', 'z_src', 'n_img'])

    summary_df = pd.DataFrame() # instantiate empty dataframe for storing summary
    for i, f_name in enumerate(H0_dicts):
        lens_i = int(os.path.splitext(f_name)[0].split('h0_dict_')[1])
        # Slice meta for this lensing system
        meta_i = meta.iloc[lens_i]
        z_lens = meta_i['z_lens']
        z_src = meta_i['z_src']
        n_img = meta_i['n_img']
        # Read in H0 samples using lens identifier
        H0_dict = np.load(os.path.join(samples_dir, f_name), allow_pickle=True).item()
        H0_samples = H0_dict['h0_samples']
        weights = H0_dict['h0_weights']
        H0_normal_stats = h0_utils.get_normal_stats_naive(H0_samples, weights)
        n_eff = np.sum(weights)**2.0/(np.sum(weights**2.0))
        # Convert H0 H0_samples to D_dt
        cosmo_converter = h0_utils.CosmoConverter(z_lens, z_src)
        D_dt_samples = cosmo_converter.get_D_dt(H0_samples)
        D_dt_stats = h0_utils.get_lognormal_stats_naive(D_dt_samples, weights)
        D_dt_normal_stats = h0_utils.get_normal_stats_naive(D_dt_samples, weights)
        summary_i = dict(
                         id=lens_i,
                         measured_td_wrt0=list(H0_dict['measured_td_wrt0']),
                         H0_mean=H0_normal_stats['mean'],
                         H0_std=H0_normal_stats['std'],
                         D_dt_mu=D_dt_stats['mu'],
                         D_dt_sigma=D_dt_stats['sigma'],
                         D_dt_mean=D_dt_normal_stats['mean'],
                         D_dt_std=D_dt_normal_stats['std'],
                         n_eff=n_eff,
                         z_lens=z_lens,
                         z_src=z_src,
                         n_img=n_img,
                         inference_time=H0_dict['inference_time'],
                         )
        summary_df = summary_df.append(summary_i, ignore_index=True)
    summary_df.to_csv(os.path.join(samples_dir, '..', 'summary.csv'))
    # Output list of problem lens IDs
    problem_id = summary_df.loc[(summary_df['n_eff'] < 3) | (summary_df['H0_std'] < 1.0)]['id'].astype(int)
    with open(os.path.join(samples_dir, '..', "mcmc_default_candidates.txt"), "w") as f:
        for pid in problem_id:
            f.write(str(pid) +"\n")

def summarize_mcmc(samples_dir, test_cfg, sampling_method, rung_idx):
    """Summarize the output of mcmc_default, i.e. MCMC samples from the D_dt posterior for each lens

    """
    true_H0 = 70.0
    true_Om0 = 0.3
    if 'mcmc_default' in sampling_method:
        if rung_idx is None:
            # Read in the relevant columns of metadata, 
            baobab_cfg = BaobabConfig.from_file(test_cfg.data.test_baobab_cfg_path)
            metadata_path = os.path.join(baobab_cfg.out_dir, 'metadata.csv')
            summary_df = pd.read_csv(metadata_path, index_col=None, usecols=['z_lens', 'z_src', 'n_img'], nrows=500) # FIXME: capped test set size at 500, as the stored dataset may be much larger
        else:
            summary_df = tdlmc_utils.convert_to_dataframe(rung=rung_idx, save_csv_path=None)
            summary_df.sort_values('seed', axis=0, inplace=True)
            true_H0 = summary_df.iloc[0]['H0']
            true_Om0 = 0.27
        summary_df['id'] = summary_df.index
        summary_df['D_dt_mu'] = np.nan
        summary_df['D_dt_sigma'] = np.nan
        summary_df['H0_mean'] = np.nan
        summary_df['H0_std'] = np.nan
        summary_df['inference_time'] = 0.0
    else:
        summary_df = pd.read_csv(os.path.join(samples_dir, '..', 'summary.csv'), index_col=None) 

    D_dt_dicts = [f for f in os.listdir(samples_dir) if f.startswith('D_dt_dict')]
    D_dt_dicts.sort()
    oversampling = 20
    threshold = 1000
    # Initialize list for catastrophic lenses not solved by MCMC
    lenses_to_rerun = []
    lenses_run = []
    for i, f_name in enumerate(D_dt_dicts):
        lens_i = int(os.path.splitext(f_name)[0].split('D_dt_dict_')[1])
        lenses_run.append(lens_i)
        meta = summary_df.loc[summary_df['id']==lens_i, ['z_lens', 'z_src']].squeeze()
        # Read in D_dt samples using lens identifier
        D_dt_dict = np.load(os.path.join(samples_dir, f_name), allow_pickle=True).item()
        # Rescale D_dt samples to correct for k_ext
        uncorrected_D_dt_samples = D_dt_dict['D_dt_samples'] # [old_n_samples,]
        uncorrected_D_dt_samples = h0_utils.remove_outliers_from_lognormal(uncorrected_D_dt_samples, 3).reshape(-1, 1) # [n_samples, 1] 
        k_ext_rv = getattr(scipy.stats, test_cfg.kappa_ext_prior.dist)(**test_cfg.kappa_ext_prior.kwargs)
        k_ext = k_ext_rv.rvs(size=[len(uncorrected_D_dt_samples), oversampling]) # [n_samples, oversampling]
        if test_cfg.kappa_ext_prior.transformed:
            D_dt_samples = (uncorrected_D_dt_samples*k_ext).flatten()
        else:
            D_dt_samples = (uncorrected_D_dt_samples/(1.0 - k_ext)).flatten() # [n_samples,]
        # Compute lognormal params for D_dt and update summary
        try:
            D_dt_stats = h0_utils.get_lognormal_stats(D_dt_samples)
            D_dt_normal_stats = h0_utils.get_normal_stats(D_dt_samples)
        except:
            print("lens", lens_i)
            print("==========")
            lenses_to_rerun.append(lens_i)
            #continue
        summary_df.loc[summary_df['id']==lens_i, 'D_dt_mu'] = D_dt_stats['mu']
        summary_df.loc[summary_df['id']==lens_i, 'D_dt_sigma'] = D_dt_stats['sigma']
        summary_df.loc[summary_df['id']==lens_i, 'D_dt_mean'] = D_dt_normal_stats['mean']
        summary_df.loc[summary_df['id']==lens_i, 'D_dt_std'] = D_dt_normal_stats['std']
        # Convert D_dt samples to H0
        D_dt_samples = scipy.stats.lognorm.rvs(scale=np.exp(D_dt_stats['mu']), s=D_dt_stats['sigma'], size=oversampling*threshold)
        D_dt_samples = D_dt_samples[np.isfinite(D_dt_samples)]
        cosmo_converter = h0_utils.CosmoConverter(meta['z_lens'], meta['z_src'], H0=true_H0, Om0=true_Om0)
        H0_samples = cosmo_converter.get_H0(D_dt_samples)
        # Reject H0 samples outside H0 prior
        H0_samples = H0_samples[np.isfinite(H0_samples)]
        if len(H0_samples) > 0:
            H0_samples = H0_samples[np.logical_and(H0_samples > 50.0, H0_samples < 90.0)]
        if len(H0_samples) < threshold:
            lenses_to_rerun.append(lens_i)
        summary_df.loc[summary_df['id']==lens_i, 'H0_mean'] = np.mean(H0_samples)
        summary_df.loc[summary_df['id']==lens_i, 'H0_std'] = np.std(H0_samples)
        summary_df.loc[summary_df['id']==lens_i, 'inference_time'] += D_dt_dict['inference_time']
    # Replace existing summary
    summary_df.to_csv(os.path.join(samples_dir, '..', 'summary.csv'))
    # Output list of catastrophic/no-good lens IDs
    if sampling_method == 'mcmc_default':
        # List of lenses that skipped MCMC
        total_lenses = np.arange(test_cfg.data.n_test)
        lenses_not_run = set(list(total_lenses)) - set(list(lenses_run))
        lenses_for_hybrid = list(lenses_not_run.union(set(lenses_to_rerun)))
        with open(os.path.join(samples_dir, '..', "hybrid_candidates.txt"), "w") as f:
            for lens_i in lenses_for_hybrid:
                f.write(str(lens_i) +"\n")
    else: # hybrid case
        with open(os.path.join(samples_dir, '..', "no_good_candidates.txt"), "w") as f:
            for lens_i in lenses_to_rerun:
                f.write(str(lens_i) +"\n")

if __name__ == '__main__':
    main()


from abc import ABC, abstractmethod
import random
import numpy as np
import torch
__all__ = ['BaseGaussianBNNPosterior', 'DiagonalGaussianBNNPosterior', 'LowRankGaussianBNNPosterior', 'DoubleLowRankGaussianBNNPosterior', 'FullRankGaussianBNNPosterior', 'DoubleGaussianBNNPosterior']

class BaseGaussianBNNPosterior(ABC):
    """Abstract base class to represent the Gaussian BNN posterior

    Gaussian posteriors or mixtures thereof with various forms of the covariance matrix inherit from this class.

    """
    def __init__(self, Y_dim, device, Y_mean=None, Y_std=None):
        """
        Parameters
        ----------
        Y_dim : int
            number of parameters to predict
        whitened_Y_cols_idx : list
            list of Y_cols indices that were whitened
        Y_mean : list
            mean values for the original values of `whitened_Y_cols`
        Y_std : list
            std values for the original values of `whitened_Y_cols`
        device : torch.device object

        """
        self.Y_dim = Y_dim
        self.Y_mean = torch.Tensor(Y_mean).reshape(1, -1)
        self.Y_std = torch.Tensor(Y_std).reshape(1, -1)
        self.device = device
        self.sigmoid = torch.nn.Sigmoid()
        self.logsigmoid = torch.nn.LogSigmoid()

    def seed_samples(self, sample_seed):
        """Seed the sampling for reproducibility

        Parameters
        ----------
        sample_seed : int

        """
        np.random.seed(sample_seed)
        random.seed(sample_seed)
        torch.manual_seed(sample_seed)
        torch.cuda.manual_seed(sample_seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False

    @abstractmethod
    def sample(self, n_samples, sample_seed=None):
        """Sample from the Gaussian posterior. Must be overridden by subclasses.

        Parameters
        ----------
        n_samples : int
            how many samples to obtain
        sample_seed : int
            seed for the samples. Default: None

        Returns
        -------
        np.array of shape `[n_samples, self.Y_dim]`
            samples

        """
        return NotImplemented

    @abstractmethod
    def get_hpd_interval(self):
        """Get the highest posterior density (HPD) interval

        """
        return NotImplemented

    def transform_back_mu(self, tensor):
        """Transform back, i.e. unwhiten, the tensor of central values

        Parameters
        ----------
        tensor : torch.Tensor of shape `[batch_size, Y_dim]`

        Returns
        -------
        torch.tensor of shape `[batch_size, Y_dim]`
            the original tensor

        """
        tensor = tensor.unsqueeze(1)
        tensor = self.unwhiten_back(tensor)
        return tensor.squeeze()

    def transform_back_logvar(self, logvar):
        """Transform back, i.e. unwhiten, the tensor of predicted log of the diagonal entries of the cov mat

        Parameters
        ----------
        tensor : torch.Tensor of shape `[batch_size, Y_dim]`

        Returns
        -------
        torch.tensor of shape `[batch_size, Y_dim]`
            the original tensor

        """
        natural_logvar = logvar*self.Y_std*self.Y_std # note self.Y_std is shape [1, Y_dim]
        return natural_logvar

    def transform_back_cov_mat(self, cov_mat):
        """Transform back, i.e. unwhiten, the tensor of predicted covariance matrix

        Parameters
        ----------
        tensor : torch.Tensor of shape `[batch_size, Y_dim, Y_dim]`

        Returns
        -------
        torch.tensor of shape `[batch_size, Y_dim]`
            the original tensor

        """
        natural_cov_mat = cov_mat*self.Y_std.unsqueeze(-1)*self.Y_std.unsqueeze(0)
        return natural_cov_mat

    def unwhiten_back(self, sample):
        """Scale and shift back to the unwhitened state

        Parameters
        ----------
        pred : torch.Tensor
            network prediction of shape `[batch_size, n_samples, self.Y_dim]`

        Returns
        -------
        torch.Tensor
            the unwhitened pred
        
        """
        sample = sample*self.Y_std.unsqueeze(1) + self.Y_mean.unsqueeze(1)
        return sample

    def sample_low_rank(self, n_samples, mu, logvar, F):
        """Sample from a single Gaussian posterior with a full but low-rank plus diagonal covariance matrix

        Parameters
        ----------
        n_samples : int
            how many samples to obtain
        mu : torch.Tensor of shape `[self.batch_size, self.Y_dim]`
            network prediction of the mu (mean parameter) of the BNN posterior
        logvar : torch.Tensor of shape `[self.batch_size, self.Y_dim]`
            network prediction of the log of the diagonal elements of the covariance matrix
        F : torch.Tensor of shape `[self.batch_size, self.Y_dim, self.rank]`
            network prediction of the low rank portion of the covariance matrix

        Returns
        -------
        np.array of shape `[self.batch_size, n_samples, self.Y_dim]`
            samples

        """
        #F = torch.unsqueeze(F, dim=1).repeat(1, n_samples, 1, 1) # [self.batch_size, n_samples, self.Y_dim, self.rank]
        F = F.repeat(n_samples, 1, 1) # [self.batch_size*n_samples, self.Y_dim, self.rank]
        mu = mu.repeat(n_samples, 1) # [self.batch_size*n_samples, self.Y_dim]
        logvar = logvar.repeat(n_samples, 1) # [self.batch_size*n_samples, self.Y_dim]
        eps_low_rank = torch.randn(self.batch_size*n_samples, self.rank, 1)
        eps_diag = torch.randn(self.batch_size*n_samples, self.Y_dim)
        half_var = torch.exp(0.5*logvar) # [self.batch_size*n_samples, self.Y_dim]
        samples = torch.bmm(F, eps_low_rank).squeeze() + mu + half_var*eps_diag
        samples = samples.reshape(n_samples, self.batch_size, self.Y_dim)
        samples = samples.transpose(0, 1)
        samples = self.unwhiten_back(samples)
        samples = samples.data.cpu().numpy()
        return samples

    def sample_full_rank(self, n_samples, mu, tril_elements, as_numpy=True):
        """Sample from a single Gaussian posterior with a full-rank covariance matrix

        Parameters
        ----------
        n_samples : int
            how many samples to obtain
        mu : torch.Tensor of shape `[self.batch_size, self.Y_dim]`
            network prediction of the mu (mean parameter) of the BNN posterior
        tril_elements : torch.Tensor of shape `[self.batch_size, tril_len]`
            network prediction of lower-triangular matrix in the log-Cholesky decomposition of the precision matrix

        Returns
        -------
        np.array of shape `[self.batch_size, n_samples, self.Y_dim]`
            samples

        """
        samples = torch.zeros([self.batch_size, n_samples, self.Y_dim])
        for b in range(self.batch_size):
            tril = torch.zeros([self.Y_dim, self.Y_dim], device=self.device, dtype=None)
            tril[self.tril_idx[0], self.tril_idx[1]] = tril_elements[b, :]
            log_diag_tril = torch.diagonal(tril, offset=0, dim1=0, dim2=1)
            tril[torch.eye(self.Y_dim, dtype=bool)] = torch.exp(log_diag_tril)
            prec_mat = torch.mm(tril, tril.T) # [Y_dim, Y_dim]
            mvn = torch.distributions.multivariate_normal.MultivariateNormal(loc=mu[b, :], precision_matrix=prec_mat)
            sample_b = mvn.sample([n_samples,])
            samples[b, :, :] = sample_b
        samples = self.unwhiten_back(samples)
        if as_numpy:
            return samples.cpu().numpy()
        else:
            return samples

class DiagonalGaussianBNNPosterior(BaseGaussianBNNPosterior):
    """The negative log likelihood (NLL) for a single Gaussian with diagonal covariance matrix
        
    `BaseGaussianNLL.__init__` docstring for the parameter description.

    """
    def __init__(self, Y_dim, device, Y_mean=None, Y_std=None):
        super(DiagonalGaussianBNNPosterior, self).__init__(Y_dim, device, Y_mean, Y_std)
        self.out_dim = self.Y_dim*2

    def set_sliced_pred(self, pred):
        d = self.Y_dim # for readability
        self.batch_size = pred.shape[0]
        self.mu = pred[:, :d]
        self.logvar = pred[:, d:]
        self.cov_diag = torch.exp(self.logvar)

    def sample(self, n_samples, sample_seed):
        """Sample from a Gaussian posterior with diagonal covariance matrix

        Parameters
        ----------
        n_samples : int
            how many samples to obtain
        sample_seed : int
            seed for the samples. Default: None

        Returns
        -------
        np.array of shape `[n_samples, self.Y_dim]`
            samples

        """
        self.seed_samples(sample_seed)
        eps = torch.randn(self.batch_size, n_samples, self.Y_dim)
        samples = eps*torch.exp(0.5*self.logvar.unsqueeze(1)) + self.mu.unsqueeze(1)
        samples = self.unwhiten_back(samples)
        samples = samples.data.cpu().numpy()
        return samples

    def get_hpd_interval(self):
        return NotImplementedError

class LowRankGaussianBNNPosterior(BaseGaussianBNNPosterior):
    """The negative log likelihood (NLL) for a single Gaussian with diagonal covariance matrix
        
    `BaseGaussianNLL.__init__` docstring for the parameter description.

    """
    def __init__(self, Y_dim, device, Y_mean=None, Y_std=None):
        super(LowRankGaussianBNNPosterior, self).__init__(Y_dim, device, Y_mean, Y_std)
        self.out_dim = self.Y_dim*4
        self.rank = 2 # FIXME: hardcoded

    def set_sliced_pred(self, pred):
        d = self.Y_dim # for readability
        self.batch_size = pred.shape[0]
        self.mu = pred[:, :d]
        self.logvar = pred[:, d:2*d]
        self.F = pred[:, 2*d:].reshape([self.batch_size, self.Y_dim, self.rank])
        F_F_tran = torch.bmm(self.F, torch.transpose(self.F, 1, 2)) # [n_lenses, d, d]
        self.cov_diag = torch.exp(self.logvar) + torch.diagonal(F_F_tran, dim1=1, dim2=2) # [n_lenses, d]
        self.cov_mat = torch.diag_embed(self.logvar) + F_F_tran

    def sample(self, n_samples, sample_seed):
        self.seed_samples(sample_seed)
        return self.sample_low_rank(n_samples, self.mu, self.logvar, self.F)

    def get_hpd_interval(self):
        return NotImplementedError

class DoubleLowRankGaussianBNNPosterior(BaseGaussianBNNPosterior):
    """The negative log likelihood (NLL) for a single Gaussian with diagonal covariance matrix
        
    `BaseGaussianNLL.__init__` docstring for the parameter description.

    """
    def __init__(self, Y_dim, device, Y_mean=None, Y_std=None):
        super(DoubleLowRankGaussianBNNPosterior, self).__init__(Y_dim, device, Y_mean, Y_std)
        self.out_dim = self.Y_dim*8 + 1
        self.rank = 2 # FIXME: hardcoded

    def set_sliced_pred(self, pred):
        d = self.Y_dim # for readability
        self.batch_size = pred.shape[0]
        # First gaussian
        self.mu = pred[:, :d]
        self.logvar = pred[:, d:2*d]
        self.F = pred[:, 2*d:4*d].reshape([self.batch_size, self.Y_dim, self.rank])
        F_F_tran = torch.bmm(self.F, torch.transpose(self.F, 1, 2)) # [n_lenses, d, d]
        self.cov_diag = torch.exp(self.logvar) + torch.diagonal(F_F_tran, dim1=1, dim2=2) # [n_lenses, d]
        self.cov_mat = torch.diag_embed(self.logvar) + F_F_tran
        # Second gaussian
        self.mu2 = pred[:, 4*d:5*d]
        self.logvar2 = pred[:, 5*d:6*d]
        self.F2 = pred[:, 6*d:8*d].reshape([self.batch_size, self.Y_dim, self.rank])
        F_F_tran2 = torch.bmm(self.F2, torch.transpose(self.F2, 1, 2))
        self.cov_diag2 = torch.exp(self.logvar2) + torch.diagonal(F_F_tran2, dim1=1, dim2=2)
        self.cov_mat2 = torch.diag_embed(self.logvar2) + F_F_tran2
        self.w2 = 0.5*self.sigmoid(pred[:, -1].reshape(-1, 1))
        
    def sample(self, n_samples, sample_seed):
        """Sample from a mixture of two Gaussians, each with a full but constrained as low-rank plus diagonal covariance

        Parameters
        ----------
        n_samples : int
            how many samples to obtain
        sample_seed : int
            seed for the samples. Default: None

        Returns
        -------
        np.array of shape `[self.batch_size, n_samples, self.Y_dim]`
            samples

        """
        self.seed_samples(sample_seed)
        samples = torch.zeros([self.batch_size, n_samples, self.Y_dim], device=self.device)
        # Determine first vs. second Gaussian
        unif2 = torch.rand(self.batch_size, n_samples)
        second_gaussian = (self.w2 > unif2)
        # Sample from second Gaussian
        samples2 = torch.Tensor(self.sample_low_rank(n_samples, self.mu2, self.logvar2, self.F2))
        samples[second_gaussian, :] = samples2[second_gaussian, :]
        # Sample from first Gaussian
        samples1 = torch.Tensor(self.sample_low_rank(n_samples, self.mu, self.logvar, self.F))
        samples[~second_gaussian, :] = samples1[~second_gaussian, :]
        samples = samples.data.cpu().numpy()
        return samples

    def get_hpd_interval(self):
        return NotImplementedError

class FullRankGaussianBNNPosterior(BaseGaussianBNNPosterior):
    """The negative log likelihood (NLL) for a single Gaussian with diagonal covariance matrix
        
    `BaseGaussianNLL.__init__` docstring for the parameter description.

    """
    def __init__(self, Y_dim, device, Y_mean=None, Y_std=None):
        super(FullRankGaussianBNNPosterior, self).__init__(Y_dim, device, Y_mean, Y_std)
        self.tril_idx = torch.tril_indices(self.Y_dim, self.Y_dim, offset=0, device=device) # lower-triangular indices
        self.tril_len = len(self.tril_idx[0])
        self.out_dim = self.Y_dim + self.Y_dim*(self.Y_dim + 1)//2

    def set_sliced_pred(self, pred):
        d = self.Y_dim # for readability
        self.batch_size = pred.shape[0]
        self.mu = pred[:, :d]
        self.tril_elements = pred[:, d:self.out_dim]

    def sample(self, n_samples, sample_seed):
        self.seed_samples(sample_seed)
        return self.sample_full_rank(n_samples, self.mu, self.tril_elements)

    def get_hpd_interval(self):
        return NotImplementedError

class DoubleGaussianBNNPosterior(BaseGaussianBNNPosterior):
    """The negative log likelihood (NLL) for a single Gaussian with diagonal covariance matrix
        
    `BaseGaussianNLL.__init__` docstring for the parameter description.

    """
    def __init__(self, Y_dim, device, Y_mean=None, Y_std=None):
        super(DoubleGaussianBNNPosterior, self).__init__(Y_dim, device, Y_mean, Y_std)
        self.tril_idx = torch.tril_indices(self.Y_dim, self.Y_dim, offset=0, device=device) # lower-triangular indices
        self.tril_len = len(self.tril_idx[0])
        self.out_dim = self.Y_dim**2 + 3*self.Y_dim + 1

    def set_sliced_pred(self, pred):
        d = self.Y_dim # for readability
        self.batch_size = pred.shape[0]
        # First gaussian
        self.mu = pred[:, :d]
        self.tril_elements = pred[:, d:d+self.tril_len]
        self.mu2 = pred[:, d+self.tril_len:2*d+self.tril_len]
        self.tril_elements2 = pred[:, 2*d+self.tril_len:-1]
        #print(pred[:, -1])
        self.w2 = 0.5*self.sigmoid(pred[:, -1].reshape(-1, 1))
        
    def sample(self, n_samples, sample_seed):
        """Sample from a mixture of two Gaussians, each with a full but constrained as low-rank plus diagonal covariance

        Parameters
        ----------
        n_samples : int
            how many samples to obtain
        sample_seed : int
            seed for the samples. Default: None

        Returns
        -------
        np.array of shape `[self.batch_size, n_samples, self.Y_dim]`
            samples

        """
        self.seed_samples(sample_seed)
        samples = torch.zeros([self.batch_size, n_samples, self.Y_dim], device=self.device)
        # Determine first vs. second Gaussian
        unif2 = torch.rand(self.batch_size, n_samples)
        second_gaussian = (self.w2 > unif2)
        # Sample from second Gaussian
        samples2 = self.sample_full_rank(n_samples, self.mu2, self.tril_elements2, as_numpy=False)
        samples[second_gaussian, :] = samples2[second_gaussian, :]
        # Sample from first Gaussian
        samples1 = self.sample_full_rank(n_samples, self.mu, self.tril_elements, as_numpy=False)
        samples[~second_gaussian, :] = samples1[~second_gaussian, :]
        samples = samples.data.cpu().numpy()
        return samples

    def get_hpd_interval(self):
        return NotImplementedError


from abc import ABC, abstractmethod
import numpy as np
import torch
from torch.distributions.multivariate_normal import MultivariateNormal
from torch.distributions.lowrank_multivariate_normal import LowRankMultivariateNormal
__all__ = ['BaseGaussianNLLNative', 'DiagonalGaussianNLLNative', 'LowRankGaussianNLLNative', 'DoubleLowRankGaussianNLLNative', 'FullRankGaussianNLLNative', 'DoubleGaussianNLLNative']

log_2_pi = 1.8378770664093453
log_2 = 0.6931471805599453

class BaseGaussianNLLNative(ABC):
    """Abstract base class to represent the Gaussian negative log likelihood (NLLNative).

    Gaussian NLLNatives or mixtures thereof with various forms of the covariance matrix inherit from this class.

    """
    def __init__(self, Y_dim, device):
        """
        Parameters
        ----------
        Y_dim : int
            number of parameters to predict
        device : torch.device object

        """
        self.Y_dim = Y_dim
        self.device = device
        self.sigmoid = torch.nn.Sigmoid()
        self.logsigmoid = torch.nn.LogSigmoid()

    @abstractmethod
    def slice(self, pred):
        """Slice the raw network prediction into meaningful Gaussian parameters

        Parameters
        ----------
        pred : torch.Tensor of shape `[batch_size, self.Y_dim]`
            the network prediction

        """
        return NotImplemented

    @abstractmethod
    def __call__(self, pred, target):
        """Evaluate the NLLNative. Must be overridden by subclasses.

        Parameters
        ----------
        pred : torch.Tensor
            raw network output for the predictions
        target : torch.Tensor
            Y labels

        """
        return NotImplemented

    def nll_diagonal(self, target, mu, logvar):
        """Evaluate the NLLNative for single Gaussian with diagonal covariance matrix

        Parameters
        ----------
        target : torch.Tensor of shape [batch_size, Y_dim]
            Y labels
        mu : torch.Tensor of shape [batch_size, Y_dim]
            network prediction of the mu (mean parameter) of the BNN posterior
        logvar : torch.Tensor of shape [batch_size, Y_dim]
            network prediction of the log of the diagonal elements of the covariance matrix

        Returns
        -------
        torch.Tensor of shape
            NLL values

        """
        precision = torch.exp(-logvar)
        # Loss kernel
        loss = precision * (target - mu)**2.0 + logvar
        # Restore prefactors
        loss += np.log(2.0*np.pi)
        loss *= 0.5
        return torch.mean(torch.sum(loss, dim=1), dim=0)

    def nll_low_rank(self, target, mu, logvar, F, reduce=True):
        """Evaluate the NLLNative for a single Gaussian with a full but low-rank plus diagonal covariance matrix

        Parameters
        ----------
        target : torch.Tensor of shape [batch_size, Y_dim]
            Y labels
        mu : torch.Tensor of shape [batch_size, Y_dim]
            network prediction of the mu (mean parameter) of the BNN posterior
        logvar : torch.Tensor of shape [batch_size, Y_dim]
            network prediction of the log of the diagonal elements of the covariance matrix
        F : torch.Tensor of shape [batch_size, rank*Y_dim]
            network prediction of the low rank portion of the covariance matrix
        reduce : bool
            whether to take the mean across the batch

        Returns
        -------
        torch.Tensor of shape [batch_size,]
            NLL values

        """
        # 1/(Y_dim - 1) * (sq_mahalanobis + log(det of \Sigma))
        batch_size, _ = target.shape # self.Y_dim = Y_dim - 1
        rank = 2 # FIXME: hardcoded for rank 2
        F = F.reshape([batch_size, self.Y_dim, rank]) 
        lr_mvn = LowRankMultivariateNormal(loc=mu, cov_factor=F, cov_diag=torch.exp(logvar))
        loss = -lr_mvn.log_prob(target)

        if reduce==True:
            return torch.mean(loss, dim=0) # float
        else:
            return loss # [batch_size,]

    def nll_mixture_low_rank(self, target, mu, logvar, F, mu2, logvar2, F2, alpha):
        """Evaluate the NLLNative for a single Gaussian with a full but low-rank plus diagonal covariance matrix
        Parameters
        ----------
        target : torch.Tensor of shape [batch_size, Y_dim]
            Y labels
        mu : torch.Tensor of shape [batch_size, Y_dim]
            network prediction of the mu (mean parameter) of the BNN posterior for the first Gaussian
        logvar : torch.Tensor of shape [batch_size, Y_dim]
            network prediction of the log of the diagonal elements of the covariance matrix for the first Gaussian
        F : torch.Tensor of shape [batch_size, rank*Y_dim]
            network prediction of the low rank portion of the covariance matrix for the first Gaussian
        mu2 : torch.Tensor of shape [batch_size, Y_dim]
            network prediction of the mu (mean parameter) of the BNN posterior for the second Gaussian
        logvar2 : torch.Tensor of shape [batch_size, Y_dim]
            network prediction of the log of the diagonal elements of the covariance matrix for the second Gaussian
        F2 : torch.Tensor of shape [batch_size, rank*Y_dim]
            network prediction of the low rank portion of the covariance matrix for the second Gaussian
        alpha : torch.Tensor of shape [batch_size, 1]
            network prediction of the logit of twice the weight on the second Gaussian 
        reduce : bool
            whether to take the mean across the batch
        Note
        ----
        The weight on the second Gaussian is required to be less than 0.5, to make the two Gaussians well-defined.
        Returns
        -------
        torch.Tensor of shape [batch_size,]
            NLL values
        """
        batch_size, _ = target.shape
        alpha = alpha.reshape(-1)
        #log_w1p1 = -alpha -torch.log1p(torch.exp(-alpha)) - self.nll_low_rank(target, mu, logvar, F=F, reduce=False) # [batch_size]
        #log_w2p2 = self.logsigmoid(alpha) - self.nll_low_rank(target, mu2, logvar2, F=F2, reduce=False) # [batch_size], 0.6931471 = np.log(2)
        log_w1p1 = torch.log1p(2.0*torch.exp(-alpha)) - log_2 - torch.log1p(torch.exp(-alpha)) - self.nll_low_rank(target, mu, logvar, F=F, reduce=False)
        log_w2p2 = -log_2 + self.logsigmoid(alpha) - self.nll_low_rank(target, mu2, logvar2, F=F2, reduce=False)
        stacked = torch.stack([log_w1p1, log_w2p2], dim=1)
        log_nll = -torch.logsumexp(stacked, dim=1)
        return torch.mean(log_nll)

    def nll_full_rank(self, target, mu, tril_elements, reduce=True):
        """Evaluate the NLLNative for a single Gaussian with a full-rank covariance matrix

        Parameters
        ----------
        target : torch.Tensor of shape [batch_size, Y_dim]
            Y labels
        mu : torch.Tensor of shape [batch_size, Y_dim]
            network prediction of the mu (mean parameter) of the BNN posterior
        tril_elements : torch.Tensor of shape [batch_size, Y_dim*(Y_dim + 1)//2]
        reduce : bool
            whether to take the mean across the batch

        Returns
        -------
        torch.Tensor of shape [batch_size,]
            NLL values

        """
        batch_size, _ = target.shape
        tril = torch.zeros([batch_size, self.Y_dim, self.Y_dim], device=self.device, dtype=None)
        tril[:, self.tril_idx[0], self.tril_idx[1]] = tril_elements
        log_diag_tril = torch.diagonal(tril, offset=0, dim1=1, dim2=2) # [batch_size, Y_dim]
        tril[:, torch.eye(self.Y_dim, dtype=bool)] = torch.exp(log_diag_tril)
        prec_mat = torch.bmm(tril, torch.transpose(tril, 1, 2))
        mvn = MultivariateNormal(loc=mu, precision_matrix=prec_mat)
        loss = -mvn.log_prob(target)
        if reduce:
            return torch.mean(loss, dim=0) # float
        else:
            return loss # [batch_size,]

    def nll_mixture(self, target, mu, tril_elements, mu2, tril_elements2, alpha):
        """Evaluate the NLLNative for a single Gaussian with a full but low-rank plus diagonal covariance matrix

        Parameters
        ----------
        target : torch.Tensor of shape [batch_size, Y_dim]
            Y labels
        mu : torch.Tensor of shape [batch_size, Y_dim]
            network prediction of the mu (mean parameter) of the BNN posterior for the first Gaussian
        tril_elements : torch.Tensor of shape [batch_size, self.tril_len]
            network prediction of the elements in the precision matrix
        mu2 : torch.Tensor of shape [batch_size, Y_dim]
            network prediction of the mu (mean parameter) of the BNN posterior for the second Gaussian
        tril_elements2 : torch.Tensor of shape [batch_size, self.tril_len]
            network prediction of the elements in the precision matrix for the second Gaussian
        alpha : torch.Tensor of shape [batch_size, 1]
            network prediction of the logit of twice the weight on the second Gaussian 

        Note
        ----
        The weight on the second Gaussian is required to be less than 0.5, to make the two Gaussians well-defined.

        Returns
        -------
        torch.Tensor of shape [batch_size,]
            NLL values

        """
        batch_size, _ = target.shape
        alpha = alpha.reshape(-1)
        log_w1p1 = torch.log1p(2.0*torch.exp(-alpha)) - log_2 - torch.log1p(torch.exp(-alpha)) - self.nll_full_rank(target, mu, tril_elements, reduce=False) # [batch_size]
        log_w2p2 = -log_2 + self.logsigmoid(alpha) - self.nll_full_rank(target, mu2, tril_elements2, reduce=False) # [batch_size]
        stacked = torch.stack([log_w1p1, log_w2p2], dim=1)
        log_nll = -torch.logsumexp(stacked, dim=1)
        return torch.mean(log_nll)

class DiagonalGaussianNLLNative(BaseGaussianNLLNative):
    """The negative log likelihood (NLLNative) for a single Gaussian with diagonal covariance matrix
        
    `BaseGaussianNLLNative.__init__` docstring for the parameter description.

    """
    posterior_name = 'DiagonalGaussianBNNPosterior'

    def __init__(self, Y_dim, device):
        super(DiagonalGaussianNLLNative, self).__init__(Y_dim, device)
        self.out_dim = Y_dim*2

    def __call__(self, pred, target):
        return self.nll_diagonal(target, *self.slice(pred))

    def slice(self, pred):
        d = self.Y_dim # for readability
        return torch.split(pred, [d, d], dim=1)

class LowRankGaussianNLLNative(BaseGaussianNLLNative):
    """The negative log likelihood (NLLNative) for a single Gaussian with a full but constrained as low-rank plus diagonal covariance matrix
        
    Only rank 2 is currently supported. `BaseGaussianNLLNative.__init__` docstring for the parameter description.

    """
    posterior_name = 'LowRankGaussianBNNPosterior'

    def __init__(self, Y_dim, device):
        super(LowRankGaussianNLLNative, self).__init__(Y_dim, device)
        self.out_dim = Y_dim*4

    def __call__(self, pred, target):
        return self.nll_low_rank(target, *self.slice(pred), reduce=True)

    def slice(self, pred):
        d = self.Y_dim # for readability
        return torch.split(pred, [d, d, 2*d], dim=1)

class FullRankGaussianNLLNative(BaseGaussianNLLNative):
    """The negative log likelihood (NLLNative) for a single Gaussian with a full-rank covariance matrix
        
    See `BaseGaussianNLLNative.__init__` docstring for the parameter description.

    """
    posterior_name = 'FullRankGaussianBNNPosterior'

    def __init__(self, Y_dim, device):
        super(FullRankGaussianNLLNative, self).__init__(Y_dim, device)
        self.tril_idx = torch.tril_indices(self.Y_dim, self.Y_dim, offset=0, device=device) # lower-triangular indices
        self.tril_len = len(self.tril_idx[0])
        self.out_dim = self.Y_dim + self.Y_dim*(self.Y_dim + 1)//2

    def __call__(self, pred, target):
        return self.nll_full_rank(target, *self.slice(pred), reduce=True)

    def slice(self, pred):
        d = self.Y_dim # for readability
        return torch.split(pred, [d, self.tril_len], dim=1)

class DoubleLowRankGaussianNLLNative(BaseGaussianNLLNative):
    """The negative log likelihood (NLLNative) for a mixture of two Gaussians, each with a full but constrained as low-rank plus diagonal covariance 
        
    Only rank 2 is currently supported. `BaseGaussianNLLNative.__init__` docstring for the parameter description.

    """
    posterior_name = 'DoubleLowRankGaussianBNNPosterior'

    def __init__(self, Y_dim, device):
        super(DoubleLowRankGaussianNLLNative, self).__init__(Y_dim, device)
        self.out_dim = Y_dim*8 + 1

    def __call__(self, pred, target):
        return self.nll_mixture_low_rank(target, *self.slice(pred))

    def slice(self, pred):
        d = self.Y_dim # for readability
        #mu, logvar, F, mu2, logvar2, F2, alpha
        return torch.split(pred, [d, d, 2*d, d, d, 2*d, 1], dim=1)

class DoubleGaussianNLLNative(BaseGaussianNLLNative):
    """The negative log likelihood (NLLNative) for a mixture of two Gaussians, each with a full but constrained as low-rank plus diagonal covariance 
        
    Only rank 2 is currently supported. `BaseGaussianNLLNative.__init__` docstring for the parameter description.

    """
    posterior_name = 'DoubleGaussianBNNPosterior'

    def __init__(self, Y_dim, device):
        super(DoubleGaussianNLLNative, self).__init__(Y_dim, device)
        self.tril_idx = torch.tril_indices(self.Y_dim, self.Y_dim, offset=0, device=device) # lower-triangular indices
        self.tril_len = len(self.tril_idx[0])
        self.out_dim = self.Y_dim**2 + 3*self.Y_dim + 1

    def __call__(self, pred, target):
        return self.nll_mixture(target, *self.slice(pred))

    def slice(self, pred):
        d = self.Y_dim # for readability
        return torch.split(pred, [d, self.tril_len, d, self.tril_len, 1], dim=1)


import os
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
import corner
from lenstronomy.Plots import chain_plot
from h0rton.h0_inference import h0_utils
from scipy.stats import norm#, median_absolute_deviation
from lenstronomy.LensModel.lens_model_extensions import LensModelExtensions
from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver
#import lenstronomy.Util.util as util
import lenstronomy.Util.simulation_util as sim_util
from lenstronomy.Data.imaging_data import ImageData
from lenstronomy.Plots import plot_util
#import scipy.ndimage as ndimage

plt.rcParams.update(plt.rcParamsDefault)
plt.rc('font', family='STIXGeneral', size=15, weight='bold')
plt.rc('xtick', labelsize='small')
plt.rc('ytick', labelsize='small')
## for Palatino and other serif fonts use:
#rc('font',**{'family':'serif','serif':['Palatino']})
#plt.rc('text', usetex=True)
plt.rc('axes', linewidth=2, titlesize='medium', labelsize='medium', labelweight='bold')
params = {'mathtext.default': 'regular' }          
plt.rcParams.update(params)

__all__ = ["plot_weighted_h0_histogram", 'plot_h0_histogram', "plot_D_dt_histogram", "plot_mcmc_corner", "gaussian", "plot_forward_modeling_comparisons"]

def gaussian(x, mean, standard_deviation, amplitude):
    return amplitude * np.exp( - ((x - mean) / standard_deviation) ** 2)

def lognormal(x, mu, sig):
    return np.exp(-0.5*(np.log(x) - mu)**2.0/sig**2.0)/(x*sig*(2.0*np.pi)**0.5)

def plot_weighted_h0_histogram(all_samples, all_weights, lens_i=0, true_h0=None, include_fit_gaussian=True, save_dir='.'):
    """Plot the histogram of H0 samples, overlaid with a Gaussian fit and truth H0

    all_samples : np.array
        H0 samples
    all_weights : np.array
        H0 weights corresponding to `all_samples`, possibly including nan values

    """
    stats = h0_utils.get_normal_stats_naive(all_samples, all_weights)
    _ = plt.hist(stats['samples'], weights=stats['weights'], bins=290, alpha=0.5, density=True, edgecolor='k', color='tab:blue', range=[10.0, 300.0])
    #print(mean, std)
    x_interval_for_fit = np.linspace(10, 300, 1000) 
    # Overlay the fit gaussian pdf
    plt.plot(x_interval_for_fit, norm.pdf(x_interval_for_fit, stats['mean'], stats['std']), color='k', label='fit: mu={:0.1f}, sig={:0.1f}'.format(stats['mean'], stats['std']))
    if save_dir is not None:
        if true_h0 is not None:
            plt.axvline(x=true_h0, linestyle='--', color='red', label='truth')
        plt.xlabel('H0 (km/Mpc/s)')
        plt.ylabel('density')
        plt.title('H0 posterior for lens {0:04d}'.format(lens_i))
        plt.legend()
        save_path = os.path.join(save_dir, 'h0_histogram_{0:04d}.png'.format(lens_i))
        plt.savefig(save_path)
        plt.close()
    return stats

def plot_weighted_D_dt_histogram(all_samples, all_weights, lens_i=0, true_D_dt=None, save_dir='.'):
    """Plot the histogram of H0 samples, overlaid with a Gaussian fit and truth H0

    all_samples : np.array
        H0 samples
    all_weights : np.array
        H0 weights corresponding to `all_samples`, possibly including nan values

    """
    # Normalize weights to unity
    is_nan_mask = np.logical_or(np.isnan(all_weights), ~np.isfinite(all_weights))
    all_weights[~is_nan_mask] = all_weights[~is_nan_mask]/np.sum(all_weights[~is_nan_mask])
    samples = all_samples[~is_nan_mask]
    weights = all_weights[~is_nan_mask]
    bin_heights, bin_borders, _ = plt.hist(samples, weights=weights, bins=200, alpha=0.5, density=True, edgecolor='k', color='tab:blue', range=[0.0, 15000.0])
    bin_centers = bin_borders[:-1] + np.diff(bin_borders) / 2
    
    # Compute the weighted mean and std analytically
    lognorm_stats = h0_utils.get_lognormal_stats_naive(samples, weights)
    mu = lognorm_stats['mu']
    sigma = lognorm_stats['sigma']
    mode = lognorm_stats['mode']
    std = lognorm_stats['std']
    popt = [mu, sigma]
    #x_interval_for_fit = np.linspace(bin_borders[0], bin_borders[-1], 10000)
    x_interval_for_fit = np.linspace(bin_centers[0], bin_centers[-1], 1000) 
    # Overlay the fit gaussian pdf
    plt.plot(x_interval_for_fit, lognormal(x_interval_for_fit, *popt), color='k', label='fit: mode={:0.1f}, std={:0.1f}'.format(mode, std))
    if save_dir is not None:
        if true_D_dt is not None:
            plt.axvline(x=true_D_dt, linestyle='--', color='red', label='truth')
        plt.xlabel('D_dt (Mpc)')
        plt.ylabel('density')
        plt.title('D_dt posterior for lens {0:04d}'.format(lens_i))
        plt.legend()
        save_path = os.path.join(save_dir, 'D_dt_histogram_{0:04d}.png'.format(lens_i))
        plt.savefig(save_path)
        plt.close()
    return mu, sigma

def plot_h0_histogram(samples, lens_i=0, true_h0=None, include_fit_gaussian=True, save_dir='.'):
    """Plot the histogram of H0 samples, overlaid with a Gaussian fit and truth H0

    all_samples : np.array
        H0 samples
    all_weights : np.array
        H0 weights corresponding to `all_samples`, possibly including nan values

    """
    # Normalize weights to unity
    bin_heights, bin_borders, _ = plt.hist(samples, bins=80, alpha=0.5, density=True, edgecolor='k', color='tab:blue', range=[40.0, 100.0])
    bin_centers = bin_borders[:-1] + np.diff(bin_borders) / 2
    if include_fit_gaussian:
        # Fit a gaussian
        best_guess_mean = bin_centers[np.argmax(bin_heights)]
        popt, _ = curve_fit(gaussian, bin_centers, bin_heights, p0=[best_guess_mean, 0.3, 3.0], maxfev=10000)
        mean = popt[0]
        std = popt[1]
    else:
        # Compute the weighted mean and std analytically
        mean = np.median(samples)
        std = np.median_absolute_deviation(samples, axis=None)
        #print(mean, std)
        popt = [mean, std, 1.0/std/np.sqrt(2*np.pi)]
    #x_interval_for_fit = np.linspace(bin_borders[0], bin_borders[-1], 10000)
    x_interval_for_fit = np.linspace(bin_centers[0], bin_centers[-1], 1000) 
    # Overlay the fit gaussian pdf
    plt.plot(x_interval_for_fit, gaussian(x_interval_for_fit, *popt), color='k', label='fit: mu={:0.1f}, sig={:0.1f}'.format(mean, std))
    #if std < 1.0:
    #    bin_heights, bin_borders, _ = plt.hist(samples, weights=weights, bins=80, alpha=0.5, density=True, edgecolor='k', color='tab:blue', range=[mean - 5, mean + 5])
    #    bin_centers = bin_borders[:-1] + np.diff(bin_borders) / 2
    #    best_guess_mean = bin_centers[np.argmax(bin_heights)]
    #    popt, _ = curve_fit(gaussian, bin_centers, bin_heights, p0=[mean, 0.3, 1.0], maxfev=10000)
    #    mean = popt[0]
    #    std = popt[-1]
    #print(popt)
    if save_dir is not None:
        if true_h0 is not None:
            plt.axvline(x=true_h0, linestyle='--', color='red', label='truth')
        plt.xlabel('H0 (km/Mpc/s)')
        plt.ylabel('density')
        plt.title('H0 posterior for lens {0:04d}'.format(lens_i))
        plt.legend()
        save_path = os.path.join(save_dir, 'h0_histogram_{0:04d}.png'.format(lens_i))
        plt.savefig(save_path)
        plt.close()
    return mean, std

def plot_D_dt_histogram(all_samples, lens_i=0, true_D_dt=None, save_dir='.'):
    """Plot the histogram of D_dt samples, overlaid with a Gaussian fit and truth D_dt

    all_samples : np.array
        D_dt MCMC samples

    """
    bin_heights, bin_borders, _ = plt.hist(all_samples, bins=200, alpha=0.5, density=True, edgecolor='k', color='tab:blue', range=[0.0, 15000.0])
    bin_centers = bin_borders[:-1] + np.diff(bin_borders) / 2
    
    # Compute the mode and std for lognormal
    lognorm_stats = h0_utils.get_lognormal_stats(all_samples)
    mu = lognorm_stats['mu']
    sigma = lognorm_stats['sigma']
    mode = lognorm_stats['mode']
    std = lognorm_stats['std']
    popt = [mu, sigma]

    #x_interval_for_fit = np.linspace(bin_borders[0], bin_borders[-1], 10000)
    x_interval_for_fit = np.linspace(bin_centers[0], bin_centers[-1], 1000) 
    # Overlay the fit gaussian pdf
    plt.plot(x_interval_for_fit, lognormal(x_interval_for_fit, *popt), color='k', label='fit: mode={:0.1f}, std={:0.1f}'.format(mode, std))
    if save_dir is not None:
        if true_D_dt is not None:
            plt.axvline(x=true_D_dt, linestyle='--', color='red', label='truth')
        plt.xlabel(r'$D_{{\Delta t}}$ (Mpc)')
        plt.ylabel('density')
        plt.title(r'$D_{{\Delta t}}$ posterior for lens {0:04d}'.format(lens_i))
        plt.legend()
        save_path = os.path.join(save_dir, 'D_dt_histogram_{0:04d}.png'.format(lens_i))
        plt.savefig(save_path)
        plt.close()
    return mu, sigma

def plot_mcmc_chain(chain_list_mcmc, save_path):
    fig, ax = chain_plot.plot_chain_list(chain_list_mcmc)
    fig.savefig(save_path, dpi=100)
    plt.close()

def plot_mcmc_corner(mcmc_samples, truth, col_labels, save_path):
    fig = corner.corner(mcmc_samples, 
                        truths=truth, 
                        truth_color='r', 
                        labels=col_labels, 
                        smooth=1.0,
                        no_fill_contours=True,
                        plot_datapoints=False,
                        show_titles=True, 
                        quiet=True,
                        plot_contours=True,
                        use_math_text=True,
                        contour_kwargs=dict(linestyles='solid'),
                        levels=[0.68, 0.95],)
    fig.savefig(save_path, dpi=100)
    plt.close()

def plot_forward_modeling_comparisons(model_plot_instance, out_dir):
    """Plot the data vs. model comparisons using the Lenstronomy modelPlot tool

    Parameters
    ----------
    model_plot_instance : lenstronomy.Plots.model_plot.ModelPlot object
    out_dir : directory in which the plots will be saved

    """
    f, axes = model_plot_instance.plot_main()
    f.savefig(os.path.join(out_dir, 'main_plot_lenstronomy.png'))
    plt.close()
    f, axes = model_plot_instance.plot_separate()
    f.savefig(os.path.join(out_dir, 'separate_plot_lenstronomy.png'))
    plt.close()
    f, axes = model_plot_instance.plot_subtract_from_data_all()
    f.savefig(os.path.join(out_dir, 'subtract_plot_lenstronomy.png'))
    plt.close('all')

#TODO define coordinate grid beforehand, e.g. kwargs_data
def lens_model_plot_custom(image, ax, lensModel, kwargs_lens, numPix=500, deltaPix=0.01, sourcePos_x=0, sourcePos_y=0, point_source=False, with_caustics=False):
    """
    plots a lens model (convergence) and the critical curves and caustics

    :param ax:
    :param kwargs_lens:
    :param numPix:
    :param deltaPix:
    :return:
    """
    kwargs_data = sim_util.data_configure_simple(numPix, deltaPix)
    data = ImageData(**kwargs_data)
    _coords = data
    _frame_size = numPix * deltaPix
    x_grid, y_grid = data.pixel_coordinates
    lensModelExt = LensModelExtensions(lensModel)
    #ra_crit_list, dec_crit_list, ra_caustic_list, dec_caustic_list = lensModelExt.critical_curve_caustics(
    #    kwargs_lens, compute_window=_frame_size, grid_scale=deltaPix/2.)
    #x_grid1d = util.image2array(x_grid)
    #y_grid1d = util.image2array(y_grid)
    #kappa_result = lensModel.kappa(x_grid1d, y_grid1d, kwargs_lens)
    #kappa_result = util.array2image(kappa_result)
    #im = ax.matshow(np.log10(kappa_result), origin='lower', extent=[0, _frame_size, 0, _frame_size], cmap='Greys',vmin=-1, vmax=1) #, cmap=self._cmap, vmin=v_min, vmax=v_max)
    im = ax.matshow(image, origin='lower', extent=[0, _frame_size, 0, _frame_size])
    if with_caustics is True:
        ra_crit_list, dec_crit_list = lensModelExt.critical_curve_tiling(kwargs_lens, compute_window=_frame_size, start_scale=deltaPix, max_order=20)
        ra_caustic_list, dec_caustic_list = lensModel.ray_shooting(ra_crit_list, dec_crit_list, kwargs_lens)
        plot_util.plot_line_set(ax, _coords, ra_caustic_list, dec_caustic_list, color='orange')
        plot_util.plot_line_set(ax, _coords, ra_crit_list, dec_crit_list, color='r')
    if point_source:
        solver = LensEquationSolver(lensModel)
        theta_x, theta_y = solver.image_position_from_source(sourcePos_x, sourcePos_y, kwargs_lens,
                                                             min_distance=deltaPix, search_window=deltaPix*numPix)
        mag_images = lensModel.magnification(theta_x, theta_y, kwargs_lens)
        x_image, y_image = _coords.map_coord2pix(theta_x, theta_y)
        abc_list = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K']
        for i in range(len(x_image)):
            x_ = (x_image[i] + 0.5) * deltaPix
            y_ = (y_image[i] + 0.5) * deltaPix
            ax.plot(x_, y_, 'dk', markersize=4*(1 + np.log(np.abs(mag_images[i]))), alpha=0.5)
            ax.text(x_, y_, abc_list[i], fontsize=20, color='k')
        x_source, y_source = _coords.map_coord2pix(sourcePos_x, sourcePos_y)
        ax.plot((x_source + 0.5) * deltaPix, (y_source + 0.5) * deltaPix, '*r', markersize=5)
    #ax.plot(numPix * deltaPix*0.5 + pred['lens_mass_center_x'] + pred['src_light_center_x'], numPix * deltaPix*0.5 + pred['lens_mass_center_y'] + pred['src_light_center_y'], '*k', markersize=5)
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    ax.autoscale(False)
    return ax


import numpy as np

__all__ = ['get_goodness', 'get_precision', 'get_accuracy', 'format_submission']

def format_submission(summary):
    """Format the summary into submission form for getting the TDLMC metrics cornerplot

    """
    pass

def get_goodness(h0_means, h0_errors, true_h0):
    """Get the goodness of fit (chi square)

    Parameters
    ----------
    h0_means : np.array
        central estimate of H0 for each lens
    h0_errors : np.array
        errors corresponding to the `h0_means`
    true_h0 : np.array or float
        the true H0

    Returns
    -------
    float
        the goodness of fit metric

    """
    chi_sq = np.mean(((h0_means - true_h0)/h0_errors)**2.0)
    return chi_sq

def get_precision(h0_errors, true_h0):
    """Get the precision, i.e. how well-constrained were the estimates on average?

    Parameters
    ----------
    h0_errors : np.array
        errors corresponding to the `h0_means`
    true_h0 : np.array or float
        the true H0

    Returns
    -------
    float
        the precision metric

    """
    precision = np.mean(h0_errors/true_h0)
    return precision

def get_accuracy(h0_means, true_h0):
    """Get the accuracy, i.e. how close were the central estimates to the truth?

    Parameters
    ----------
    h0_means : np.array
        central estimate of H0 for each lens
    true_h0 : np.array or float
        the true H0

    Returns
    -------
    float
        the accuracy metric

    """
    precision = np.mean((h0_means - true_h0)/true_h0)
    return precision


"""Script to run an MCMC afterburner for the BNN posterior

It borrows heavily from the `catalogue modelling.ipynb` notebook in Lenstronomy Extensions, which you can find `here <https://github.com/sibirrer/lenstronomy_extensions/blob/master/lenstronomy_extensions/Notebooks/catalogue%20modelling.ipynb>`_.

"""
import os
import time
from tqdm import tqdm
from ast import literal_eval
import numpy as np
import pandas as pd
import torch
from torch.utils.data import DataLoader
from lenstronomy.Workflow.fitting_sequence import FittingSequence
from lenstronomy.Cosmo.lcdm import LCDM
import baobab.sim_utils.metadata_utils as metadata_utils
from h0rton.script_utils import parse_args, seed_everything, HiddenPrints
import h0rton.models
from h0rton.configs import TrainValConfig, TestConfig
import h0rton.losses
import h0rton.train_utils as train_utils
from h0rton.h0_inference import h0_utils, plotting_utils, mcmc_utils
from h0rton.trainval_data import XYCosmoData

def main():
    args = parse_args()
    test_cfg = TestConfig.from_file(args.test_config_file_path)
    train_val_cfg = TrainValConfig.from_file(test_cfg.train_val_config_file_path)
    # Set device and default data type
    device = torch.device(test_cfg.device_type)
    if device.type == 'cuda':
        torch.set_default_tensor_type('torch.cuda.FloatTensor')
    else:
        torch.set_default_tensor_type('torch.FloatTensor')
    seed_everything(test_cfg.global_seed)
    
    ############
    # Data I/O #
    ############
    test_data = XYCosmoData(test_cfg.data.test_dir, data_cfg=train_val_cfg.data)
    master_truth = test_data.cosmo_df
    master_truth = metadata_utils.add_qphi_columns(master_truth)
    master_truth = metadata_utils.add_gamma_psi_ext_columns(master_truth)
    if test_cfg.data.lens_indices is None:
        if args.lens_indices_path is None:
            # Test on all n_test lenses in the test set
            n_test = test_cfg.data.n_test 
            lens_range = range(n_test)
        else:
            # Test on the lens indices in a text file at the specified path
            lens_range = []
            with open(args.lens_indices_path, "r") as f:
                for line in f:
                    lens_range.append(int(line.strip()))
            n_test = len(lens_range)
            print("Performing H0 inference on {:d} specified lenses...".format(n_test))
    else:
        if args.lens_indices_path is None:
            # Test on the lens indices specified in the test config file
            lens_range = test_cfg.data.lens_indices
            n_test = len(lens_range)
            print("Performing H0 inference on {:d} specified lenses...".format(n_test))
        else:
            raise ValueError("Specific lens indices were specified in both the test config file and the command-line argument.")
    batch_size = max(lens_range) + 1
    test_loader = DataLoader(test_data, batch_size=batch_size, shuffle=False, drop_last=True)
    # Output directory into which the H0 histograms and H0 samples will be saved
    out_dir = test_cfg.out_dir
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
        print("Destination folder path: {:s}".format(out_dir))
    else:
        raise OSError("Destination folder already exists.")

    ######################
    # Load trained state #
    ######################
    # Instantiate loss function, to append to the MCMC objective as the prior
    orig_Y_cols = train_val_cfg.data.Y_cols
    loss_fn = getattr(h0rton.losses, train_val_cfg.model.likelihood_class)(Y_dim=train_val_cfg.data.Y_dim, device=device)
    # Instantiate MCMC parameter penalty function
    params_to_remove = ['lens_light_R_sersic'] #'src_light_R_sersic'] 
    mcmc_Y_cols = [col for col in orig_Y_cols if col not in params_to_remove]
    mcmc_Y_dim = len(mcmc_Y_cols)
    mcmc_loss_fn = getattr(h0rton.losses, train_val_cfg.model.likelihood_class)(Y_dim=train_val_cfg.data.Y_dim - len(params_to_remove), device=device)
    remove_param_idx, remove_idx = mcmc_utils.get_idx_for_params(mcmc_loss_fn.out_dim, orig_Y_cols, params_to_remove, train_val_cfg.model.likelihood_class)
    mcmc_train_Y_mean = np.delete(train_val_cfg.data.train_Y_mean, remove_param_idx)
    mcmc_train_Y_std = np.delete(train_val_cfg.data.train_Y_std, remove_param_idx)
    parameter_penalty = mcmc_utils.HybridBNNPenalty(mcmc_Y_cols, train_val_cfg.model.likelihood_class, mcmc_train_Y_mean, mcmc_train_Y_std, test_cfg.h0_posterior.exclude_velocity_dispersion, device)
    custom_logL_addition = parameter_penalty.evaluate if test_cfg.lens_posterior_type.startswith('default') else None
    null_spread = True if test_cfg.lens_posterior_type == 'truth' else False
    # Instantiate model
    net = getattr(h0rton.models, train_val_cfg.model.architecture)(num_classes=loss_fn.out_dim)
    net.to(device)
    # Load trained weights from saved state
    net, epoch = train_utils.load_state_dict_test(test_cfg.state_dict_path, net, train_val_cfg.optim.n_epochs, device)
    with torch.no_grad():
        net.eval()
        for X_, Y_ in test_loader:
            X = X_.to(device)
            Y = Y_.to(device) # TODO: compare master_truth with reverse-transformed Y
            pred = net(X)
            break

    mcmc_pred = pred.cpu().numpy()
    if test_cfg.lens_posterior_type == 'default_with_truth_mean':
        # Replace BNN posterior's primary gaussian mean with truth values
        mcmc_pred[:, :len(mcmc_Y_cols)] = Y[:, :len(mcmc_Y_cols)].cpu().numpy()
    mcmc_pred = mcmc_utils.remove_parameters_from_pred(mcmc_pred, remove_idx, return_as_tensor=False)

    # Instantiate posterior for BNN samples, to initialize the walkers
    bnn_post = getattr(h0rton.h0_inference.gaussian_bnn_posterior, loss_fn.posterior_name)(mcmc_Y_dim, device, mcmc_train_Y_mean, mcmc_train_Y_std)
    bnn_post.set_sliced_pred(torch.tensor(mcmc_pred))
    n_walkers = test_cfg.numerics.mcmc.walkerRatio*(mcmc_Y_dim + 1) # BNN params + H0 times walker ratio
    init_pos = bnn_post.sample(n_walkers, sample_seed=test_cfg.global_seed) # [batch_size, n_walkers, mcmc_Y_dim] contains just the lens model params, no D_dt
    init_D_dt = np.random.uniform(0.0, 10000.0, size=(batch_size, n_walkers, 1)) # FIXME: init H0 hardcoded

    kwargs_model = dict(lens_model_list=['PEMD', 'SHEAR'],
                        point_source_model_list=['SOURCE_POSITION'],
                        source_light_model_list=['SERSIC_ELLIPSE'])
    astro_sig = test_cfg.image_position_likelihood.sigma
    # Get H0 samples for each system
    if not test_cfg.time_delay_likelihood.baobab_time_delays:
        if 'abcd_ordering_i' not in master_truth:
            raise ValueError("If the time delay measurements were not generated using Baobab, the user must specify the order of image positions in which the time delays are listed, in order of increasing dec.")

    lenses_skipped = [] # keeps track of lenses that skipped MCMC
    total_progress = tqdm(total=n_test)
    realized_time_delays = pd.DataFrame()
    # For each lens system...
    for i, lens_i in enumerate(lens_range):
        # Each lens gets a unique random state for td and vd measurement error realizations.
        rs_lens = np.random.RandomState(lens_i)
        ###########################
        # Relevant data and prior #
        ###########################
        data_i = master_truth.iloc[lens_i].copy()
        parameter_penalty.set_bnn_post_params(mcmc_pred[lens_i, :]) # set the BNN parameters
        # Init values for the lens model params
        if test_cfg.lens_posterior_type == 'default':
            init_info = dict(zip(mcmc_Y_cols, mcmc_pred[lens_i, :len(mcmc_Y_cols)]*mcmc_train_Y_std + mcmc_train_Y_mean)) # mean of primary Gaussian
        else: # types 'hybrid_with_truth_mean' and 'truth'
            init_info = dict(zip(mcmc_Y_cols, data_i[mcmc_Y_cols].values)) # truth params
        if not test_cfg.h0_posterior.exclude_velocity_dispersion:
            parameter_penalty.set_vel_disp_params()
            raise NotImplementedError
        lcdm = LCDM(z_lens=data_i['z_lens'], z_source=data_i['z_src'], flat=True)
        true_img_dec = np.trim_zeros(data_i[['y_image_0', 'y_image_1', 'y_image_2', 'y_image_3']].values, 'b')
        n_img = len(true_img_dec)
        true_td = np.array(literal_eval(data_i['true_td']))
        measured_td = true_td + rs_lens.randn(*true_td.shape)*test_cfg.error_model.time_delay_error
        print(measured_td)
        realized_time_delays = realized_time_delays.append({'measured_td': list(measured_td)}, ignore_index=True)
        continue
        #print(true_td, measured_td)
        measured_td_sig = test_cfg.time_delay_likelihood.sigma # np.ones(n_img - 1)*
        measured_img_dec = true_img_dec + rs_lens.randn(n_img)*astro_sig
        increasing_dec_i = np.argsort(true_img_dec) #np.argsort(measured_img_dec)
        measured_td = h0_utils.reorder_to_tdlmc(measured_td, increasing_dec_i, range(n_img)) # need to use measured dec to order
        measured_img_dec = h0_utils.reorder_to_tdlmc(measured_img_dec, increasing_dec_i, range(n_img))
        measured_td_wrt0 = measured_td[1:] - measured_td[0]   
        kwargs_data_joint = dict(time_delays_measured=measured_td_wrt0,
                                 time_delays_uncertainties=measured_td_sig,
                                 #vel_disp_measured=measured_vd, # TODO: optionally exclude
                                 #vel_disp_uncertainty=vel_disp_sig,
                                 )
        if not test_cfg.h0_posterior.exclude_velocity_dispersion:
            measured_vd = data_i['true_vd']*(1.0 + rs_lens.randn()*test_cfg.error_model.velocity_dispersion_frac_error)
            kwargs_data_joint['vel_disp_measured'] = measured_vd
            kwargs_data_joint['vel_disp_sig'] = test_cfg.velocity_dispersion_likelihood.sigma

        #############################
        # Parameter init and bounds #
        #############################
        lens_kwargs = mcmc_utils.get_lens_kwargs(init_info, null_spread=null_spread)
        ps_kwargs = mcmc_utils.get_ps_kwargs_src_plane(init_info, astro_sig, null_spread=null_spread)
        src_light_kwargs = mcmc_utils.get_light_kwargs(init_info['src_light_R_sersic'], null_spread=null_spread)
        special_kwargs = mcmc_utils.get_special_kwargs(n_img, astro_sig, null_spread=null_spread) # image position offset and time delay distance, aka the "special" parameters
        kwargs_params = {'lens_model': lens_kwargs,
                         'point_source_model': ps_kwargs,
                         'source_model': src_light_kwargs,
                         'special': special_kwargs,}
        if test_cfg.numerics.solver_type == 'NONE':
            solver_type = 'NONE'
        else:
            solver_type = 'PROFILE_SHEAR' if n_img == 4 else 'CENTER'
        #solver_type = 'NONE'
        kwargs_constraints = {'num_point_source_list': [n_img],  
                              'Ddt_sampling': True,
                              'solver_type': solver_type,}

        kwargs_likelihood = {'time_delay_likelihood': True,
                             'sort_images_by_dec': True,
                             'prior_lens': [],
                             'prior_special': [],
                             'check_bounds': True, 
                             'check_matched_source_position': False,
                             'source_position_tolerance': 0.01,
                             'source_position_sigma': 0.01,
                             'source_position_likelihood': False,
                             'custom_logL_addition': custom_logL_addition,}

        ###########################
        # MCMC posterior sampling #
        ###########################
        fitting_seq = FittingSequence(kwargs_data_joint, kwargs_model, kwargs_constraints, kwargs_likelihood, kwargs_params, verbose=False, mpi=False)
        if i == 0:
            param_class = fitting_seq._updateManager.param_class
            n_params, param_class_Y_cols = param_class.num_param()
            init_pos = mcmc_utils.reorder_to_param_class(mcmc_Y_cols, param_class_Y_cols, init_pos, init_D_dt)
        # MCMC sample from the post-processed BNN posterior jointly with cosmology
        lens_i_start_time = time.time()
        if test_cfg.lens_posterior_type == 'default':
            test_cfg.numerics.mcmc.update(init_samples=init_pos[lens_i, :, :])
        fitting_kwargs_list_mcmc = [['MCMC', test_cfg.numerics.mcmc]]
        with HiddenPrints():
            try:
                chain_list_mcmc = fitting_seq.fit_sequence(fitting_kwargs_list_mcmc)
                kwargs_result_mcmc = fitting_seq.best_fit()
            except:
                print("lens {:d} skipped".format(lens_i))
                total_progress.update(1)
                lenses_skipped.append(lens_i)
                continue
        lens_i_end_time = time.time()
        inference_time = (lens_i_end_time - lens_i_start_time)/60.0 # min

        #############################
        # Plotting the MCMC samples #
        #############################
        # sampler_type : 'EMCEE'
        # samples_mcmc : np.array of shape `[n_mcmc_eval, n_params]`
        # param_mcmc : list of str of length n_params, the parameter names
        sampler_type, samples_mcmc, param_mcmc, _  = chain_list_mcmc[0]
        new_samples_mcmc = mcmc_utils.postprocess_mcmc_chain(kwargs_result_mcmc, samples_mcmc, kwargs_model, lens_kwargs[2], ps_kwargs[2], src_light_kwargs[2], special_kwargs[2], kwargs_constraints)
        # Plot D_dt histogram
        D_dt_samples = new_samples_mcmc['D_dt'].values
        true_D_dt = lcdm.D_dt(H_0=data_i['H0'], Om0=0.3)
        data_i['D_dt'] = true_D_dt
        # Export D_dt samples for this lens
        lens_inference_dict = dict(
                                   D_dt_samples=D_dt_samples, # kappa_ext=0 for these samples
                                   inference_time=inference_time,
                                   true_D_dt=true_D_dt, 
                                   )
        lens_inference_dict_save_path = os.path.join(out_dir, 'D_dt_dict_{0:04d}.npy'.format(lens_i))
        np.save(lens_inference_dict_save_path, lens_inference_dict)
        # Optionally export the MCMC samples
        if test_cfg.export.mcmc_samples:
            mcmc_samples_path = os.path.join(out_dir, 'mcmc_samples_{0:04d}.csv'.format(lens_i))
            new_samples_mcmc.to_csv(mcmc_samples_path, index=None)
        # Optionally export the D_dt histogram
        if test_cfg.export.D_dt_histogram:
            cleaned_D_dt_samples = h0_utils.remove_outliers_from_lognormal(D_dt_samples, 3)
            _ = plotting_utils.plot_D_dt_histogram(cleaned_D_dt_samples, lens_i, true_D_dt, save_dir=out_dir)
        # Optionally export the plot of MCMC chain
        if test_cfg.export.mcmc_chain:
            mcmc_chain_path = os.path.join(out_dir, 'mcmc_chain_{0:04d}.png'.format(lens_i))
            plotting_utils.plot_mcmc_chain(chain_list_mcmc, mcmc_chain_path)
        # Optionally export posterior cornerplot of select lens model parameters with D_dt
        if test_cfg.export.mcmc_corner:
            mcmc_corner_path = os.path.join(out_dir, 'mcmc_corner_{0:04d}.png'.format(lens_i))
            plotting_utils.plot_mcmc_corner(new_samples_mcmc[test_cfg.export.mcmc_cols], data_i[test_cfg.export.mcmc_cols], test_cfg.export.mcmc_col_labels, mcmc_corner_path)
        total_progress.update(1)
    realized_time_delays.to_csv(os.path.join(out_dir, 'realized_time_delays.csv'), index=False)
    total_progress.close()

if __name__ == '__main__':
    #import cProfile
    #pr = cProfile.Profile()
    #pr.enable()
    main()
    #pr.disable()
    #pr.print_stats(sort='cumtime')


import os
import shutil
import unittest
import numpy as np
import pandas as pd
from addict import Dict
from torch.utils.data import DataLoader
from h0rton.trainval_data import XYData
from baobab.configs import BaobabConfig

class TestXYData(unittest.TestCase):
    """A suite of tests on data preprocessing
    
    """
    @classmethod
    def setUpClass(cls):
        cls.Y_cols = ["lens_mass_center_x", "src_light_center_x","lens_mass_center_y", "src_light_center_y", "external_shear_gamma_ext", "external_shear_psi_ext"]
        cls.train_Y_mean = np.random.randn(len(cls.Y_cols))
        cls.train_Y_std = np.abs(np.random.randn(len(cls.Y_cols))) + 1.0
        cls.train_baobab_cfg_path = 'h0rton/tests/test_trainval_data/baobab_train.json'
        cls.val_baobab_cfg_path = 'h0rton/tests/test_trainval_data/baobab_val.json'
        cls.train_baobab_cfg = BaobabConfig.from_file(cls.train_baobab_cfg_path)
        cls.val_baobab_cfg = BaobabConfig.from_file(cls.val_baobab_cfg_path)
        cls.original_exptime = 5400.0 # value in baobab_[train/val].json

        #####################
        # Generate toy data #
        #####################
        # Training (n_data = 2)
        os.makedirs(cls.train_baobab_cfg.out_dir, exist_ok=True)
        cls.train_metadata = pd.DataFrame.from_dict({
                                                    "lens_mass_center_x": [1.5, 2.0], 
                                                    "lens_mass_center_y": [1.8, 9.0], 
                                                    "src_light_center_x": [10.1, 12.5], 
                                                    "src_light_center_y": [29.2, 18.0],
                                                    "external_shear_gamma_ext": [-0.02, 0.02],
                                                    "external_shear_psi_ext": [-0.5, 0.5], 
                                                    "img_filename": ['X_{0:07d}.npy'.format(i) for i in range(2)],
                                                    })
        cls.train_metadata.to_csv(os.path.join(cls.train_baobab_cfg.out_dir, 'metadata.csv'), index=False)
        cls.img_0 = np.abs(np.random.randn(9)*2.0).reshape([1, 3, 3])
        cls.img_1 = np.abs(np.random.randn(9)*2.0).reshape([1, 3, 3])
        np.save(os.path.join(cls.train_baobab_cfg.out_dir, 'X_{0:07d}.npy'.format(0)), cls.img_0)
        np.save(os.path.join(cls.train_baobab_cfg.out_dir, 'X_{0:07d}.npy'.format(1)), cls.img_1)
        # Validation (n_data = 3)
        os.makedirs(cls.val_baobab_cfg.out_dir, exist_ok=True)
        cls.val_metadata = pd.DataFrame.from_dict({
                                                  "lens_mass_center_x": np.random.randn(3), 
                                                  "lens_mass_center_y": np.random.randn(3), 
                                                  "src_light_center_x": np.random.randn(3), 
                                                  "src_light_center_y": np.random.randn(3), 
                                                  "external_shear_gamma_ext": np.random.randn(3),
                                                    "external_shear_psi_ext": np.random.randn(3), 
                                                  "img_filename": ['X_{0:07d}.npy'.format(i) for i in range(3)],
                                                  })
        cls.img_0_val = np.abs(np.random.randn(9)*2.0).reshape([1, 3, 3])
        cls.img_1_val = np.abs(np.random.randn(9)*2.0).reshape([1, 3, 3])
        cls.img_2_val = np.abs(np.random.randn(9)*2.0).reshape([1, 3, 3])
        np.save(os.path.join(cls.val_baobab_cfg.out_dir, 'X_{0:07d}.npy'.format(0)), cls.img_0_val)
        np.save(os.path.join(cls.val_baobab_cfg.out_dir, 'X_{0:07d}.npy'.format(1)), cls.img_1_val)
        np.save(os.path.join(cls.val_baobab_cfg.out_dir, 'X_{0:07d}.npy'.format(2)), cls.img_2_val)
        cls.val_metadata.to_csv(os.path.join(cls.val_baobab_cfg.out_dir, 'metadata.csv'), index=False)

    @classmethod
    def tearDownClass(cls):
        """Remove the toy data

        """
        shutil.rmtree(cls.train_baobab_cfg.out_dir)
        shutil.rmtree(cls.val_baobab_cfg.out_dir)

    def test_X_identity(self):
        """Test if the input iamge equals the dataset image, when nothing is done to the image at all

        """
        train_data = XYData(True, self.Y_cols, 'FloatTensor', define_src_pos_wrt_lens=True, rescale_pixels=False, log_pixels=False, add_pixel_noise=False, eff_exposure_time={'TDLMC_F160W': self.original_exptime}, train_Y_mean=self.train_Y_mean, train_Y_std=self.train_Y_std, train_baobab_cfg_path=self.train_baobab_cfg_path, val_baobab_cfg_path=self.val_baobab_cfg_path, for_cosmology=False)
        actual_img, _ = train_data[0]
        expected_img = self.img_0
        np.testing.assert_array_almost_equal(actual_img, expected_img, err_msg='test_X_identity')

    def test_X_transformation_log(self):
        """Test if the images transform as expected, with log(1+X)

        """
        train_data = XYData(True, self.Y_cols, 'FloatTensor', define_src_pos_wrt_lens=True, rescale_pixels=False, log_pixels=True, add_pixel_noise=False, eff_exposure_time={'TDLMC_F160W': self.original_exptime}, train_Y_mean=self.train_Y_mean, train_Y_std=self.train_Y_std, train_baobab_cfg_path=self.train_baobab_cfg_path, val_baobab_cfg_path=self.val_baobab_cfg_path, for_cosmology=False)
        actual_img, _ = train_data[0]
        expected_img = self.img_0
        expected_img = np.log1p(expected_img)
        np.testing.assert_array_almost_equal(actual_img, expected_img, err_msg='test_X_transformation_log')

    def test_X_transformation_rescale(self):
        """Test if the images transform as expected, with whitening

        """
        train_data = XYData(True, self.Y_cols, 'FloatTensor', define_src_pos_wrt_lens=True, rescale_pixels=True, log_pixels=False, add_pixel_noise=False, eff_exposure_time={'TDLMC_F160W': self.original_exptime}, train_Y_mean=self.train_Y_mean, train_Y_std=self.train_Y_std, train_baobab_cfg_path=self.train_baobab_cfg_path, val_baobab_cfg_path=self.val_baobab_cfg_path, for_cosmology=False)
        actual_img, _ = train_data[0]
        expected_img = self.img_0
        expected_img = (expected_img - np.mean(expected_img))/np.std(expected_img, ddof=1)
        np.testing.assert_array_almost_equal(actual_img, expected_img, err_msg='test_X_transformation_rescale')

    def test_X_transformation_log_rescale(self):
        """Test if the images transform as expected, with log(1+X) and whitening

        """
        # Without exposure time factor
        train_data = XYData(True, self.Y_cols, 'FloatTensor', define_src_pos_wrt_lens=True, rescale_pixels=True, log_pixels=True, add_pixel_noise=False, eff_exposure_time={'TDLMC_F160W': self.original_exptime}, train_Y_mean=self.train_Y_mean, train_Y_std=self.train_Y_std, train_baobab_cfg_path=self.train_baobab_cfg_path, val_baobab_cfg_path=self.val_baobab_cfg_path, for_cosmology=False)
        actual_img, _ = train_data[0]
        expected_img = self.img_0
        expected_img = np.log1p(expected_img)
        # Note torch std takes into account Bessel correction
        expected_img = (expected_img - np.mean(expected_img))/np.std(expected_img, ddof=1)
        np.testing.assert_array_almost_equal(actual_img, expected_img, err_msg='test_X_transformation_log_rescale, without exposure time factor')
        # With exposure time factor
        train_data = XYData(True, self.Y_cols, 'FloatTensor', define_src_pos_wrt_lens=True, rescale_pixels=True, log_pixels=True, add_pixel_noise=False, eff_exposure_time={'TDLMC_F160W': self.original_exptime*2.0}, train_Y_mean=self.train_Y_mean, train_Y_std=self.train_Y_std, train_baobab_cfg_path=self.train_baobab_cfg_path, val_baobab_cfg_path=self.val_baobab_cfg_path, for_cosmology=False)
        actual_img, _ = train_data[0]
        expected_img = self.img_0*2.0
        expected_img = np.log1p(expected_img)
        # Note torch std takes into account Bessel correction
        expected_img = (expected_img - np.mean(expected_img))/np.std(expected_img, ddof=1)
        np.testing.assert_array_almost_equal(actual_img, expected_img, err_msg='test_X_transformation_log_rescale, with exposure time factor')

    def test_X_exposure_time_factor(self):
        """Test if the images scale by the new effective exposure time correctly

        """
        train_data = XYData(True, self.Y_cols, 'FloatTensor', define_src_pos_wrt_lens=True, rescale_pixels=False, log_pixels=False, add_pixel_noise=False, eff_exposure_time={'TDLMC_F160W': self.original_exptime*2.0}, train_Y_mean=self.train_Y_mean, train_Y_std=self.train_Y_std, train_baobab_cfg_path=self.train_baobab_cfg_path, val_baobab_cfg_path=self.val_baobab_cfg_path, for_cosmology=False)
        actual_img, _ = train_data[0]
        expected_img = self.img_0*2.0
        np.testing.assert_array_almost_equal(actual_img, expected_img, err_msg='test_X_exposure_time_factor')

    def test_Y_transformation_(self):
        """Test if the target Y whitens correctly

        """
        # Training
        train_data = XYData(True, self.Y_cols, 'FloatTensor', define_src_pos_wrt_lens=True, rescale_pixels=False, log_pixels=False, add_pixel_noise=False, eff_exposure_time={'TDLMC_F160W': self.original_exptime*2.0}, train_Y_mean=None, train_Y_std=None, train_baobab_cfg_path=self.train_baobab_cfg_path, val_baobab_cfg_path=self.val_baobab_cfg_path, for_cosmology=False)
        _, actual_Y_0 = train_data[0]
        _, actual_Y_1 = train_data[1]
        actual_Y = np.stack([actual_Y_0, actual_Y_1], axis=0)
        Y_df = self.train_metadata[self.Y_cols].copy()
        Y_df['src_light_center_x'] -= Y_df['lens_mass_center_x']
        Y_df['src_light_center_y'] -= Y_df['lens_mass_center_y']
        expected_Y = Y_df.values
        before_whitening_Y = Y_df.values
        #expected_Y = (expected_Y - self.train_Y_mean.reshape([1, -1]))/self.train_Y_std.reshape([1, -1])
        expected_Y[np.argmin(before_whitening_Y, axis=0), np.arange(len(self.Y_cols))] = -1
        expected_Y[np.argmax(before_whitening_Y, axis=0), np.arange(len(self.Y_cols))] = 1
        np.testing.assert_array_equal(actual_Y_0.shape, [len(self.Y_cols),], err_msg='shape of single example Y for training')
        np.testing.assert_array_almost_equal(actual_Y, expected_Y, err_msg='transformed Y for training')
        # Validation
        val_data = XYData(False, self.Y_cols, 'FloatTensor', define_src_pos_wrt_lens=True, rescale_pixels=False, log_pixels=False, add_pixel_noise=False, eff_exposure_time={'TDLMC_F160W': self.original_exptime*2.0}, train_Y_mean=self.train_Y_mean, train_Y_std=self.train_Y_std, train_baobab_cfg_path=self.train_baobab_cfg_path, val_baobab_cfg_path=self.val_baobab_cfg_path, for_cosmology=False)
        _, actual_Y_0 = val_data[0]
        _, actual_Y_1 = val_data[1]
        _, actual_Y_2 = val_data[2]
        actual_Y = np.stack([actual_Y_0, actual_Y_1, actual_Y_2], axis=0)
        expected_Y = self.val_metadata[self.Y_cols].copy()
        expected_Y['src_light_center_x'] -= expected_Y['lens_mass_center_x']
        expected_Y['src_light_center_y'] -= expected_Y['lens_mass_center_y']
        expected_Y = expected_Y.values
        expected_Y = (expected_Y - self.train_Y_mean.reshape([1, -1]))/self.train_Y_std.reshape([1, -1])
        np.testing.assert_array_equal(actual_Y_0.shape, [len(self.Y_cols),], err_msg='shape of single example Y for validation for arbitrary train mean and std')
        np.testing.assert_array_almost_equal(actual_Y, expected_Y, err_msg='transformed Y for validation for arbitrary train mean and std')

    def test_train_vs_val(self):
        """Test if the images and metadata are loaded from the correct folder (train/val)

        """
        train_data = XYData(True, self.Y_cols, 'FloatTensor', define_src_pos_wrt_lens=True, rescale_pixels=False, log_pixels=False, add_pixel_noise=False, eff_exposure_time={'TDLMC_F160W': self.original_exptime*2.0}, train_Y_mean=self.train_Y_mean, train_Y_std=self.train_Y_std, train_baobab_cfg_path=self.train_baobab_cfg_path, val_baobab_cfg_path=self.val_baobab_cfg_path, for_cosmology=False)
        val_data = XYData(False, self.Y_cols, 'FloatTensor', define_src_pos_wrt_lens=True, rescale_pixels=False, log_pixels=False, add_pixel_noise=False, eff_exposure_time={'TDLMC_F160W': self.original_exptime*2.0}, train_Y_mean=self.train_Y_mean, train_Y_std=self.train_Y_std, train_baobab_cfg_path=self.train_baobab_cfg_path, val_baobab_cfg_path=self.val_baobab_cfg_path, for_cosmology=False)
        np.testing.assert_equal(len(train_data), 2, err_msg='reading from correct folder (train/val)')
        np.testing.assert_equal(len(val_data), 3, err_msg='reading from correct folder (train/val)')

    def test_tensor_type(self):
        """Test if X, Y are of the configured data type

        """
        # DoubleTensor
        train_data = XYData(True, self.Y_cols, 'DoubleTensor', define_src_pos_wrt_lens=True, rescale_pixels=False, log_pixels=False, add_pixel_noise=False, eff_exposure_time={'TDLMC_F160W': self.original_exptime*2.0}, train_Y_mean=self.train_Y_mean, train_Y_std=self.train_Y_std, train_baobab_cfg_path=self.train_baobab_cfg_path, val_baobab_cfg_path=self.val_baobab_cfg_path, for_cosmology=False)
        actual_X_0, actual_Y_0 = train_data[0]
        assert actual_X_0.type() == 'torch.DoubleTensor'
        assert actual_Y_0.type() == 'torch.DoubleTensor'

        # FloatTensor
        train_data = XYData(True, self.Y_cols, 'FloatTensor', define_src_pos_wrt_lens=True, rescale_pixels=False, log_pixels=False, add_pixel_noise=False, eff_exposure_time={'TDLMC_F160W': self.original_exptime*2.0}, train_Y_mean=self.train_Y_mean, train_Y_std=self.train_Y_std, train_baobab_cfg_path=self.train_baobab_cfg_path, val_baobab_cfg_path=self.val_baobab_cfg_path, for_cosmology=False)
        actual_X_0, actual_Y_0 = train_data[0]
        assert actual_X_0.type() == 'torch.FloatTensor'
        assert actual_Y_0.type() == 'torch.FloatTensor'

if __name__ == '__main__':
    unittest.main()


import numpy as np
from matplotlib.figure import Figure
__all__ = ['get_1d_mapping_fig', 'get_mae', 'interpret_pred', 'get_logdet']

def get_logdet(tril_elements, Y_dim):
    """Returns the log determinant of the covariance matrix

    """
    batch_size = tril_elements.shape[0]
    tril = np.zeros([batch_size, Y_dim, Y_dim])
    tril_idx = np.tril_indices(Y_dim)
    tril_len = len(tril_idx[0])
    tril[:, tril_idx[0], tril_idx[1]] = tril_elements[:, :tril_len] # safeguarding measure
    log_diag_tril = np.diagonal(tril, offset=0, axis1=1, axis2=2) # [batch_size, Y_dim]
    return -np.sum(log_diag_tril, axis=1) # [batch_size,]

def get_1d_mapping_fig(name, mu, Y):
    """Plots the marginal 1D mapping of the mean predictions

    Parameters
    ----------
    name : str
        name of the parameter
    mu : np.array of shape [batch_size,]
        network prediction of the Gaussian mean
    Y : np.array of shape [batch_size,]
        truth label
    which_normal_i : int
        which Gaussian (0 for first, 1 for second)

    Returns
    -------
    matplotlib.FigureCanvas object
        plot of network predictions against truth

    """
    my_dpi = 72.0
    fig = Figure(figsize=(720/my_dpi, 360/my_dpi), dpi=my_dpi, tight_layout=True)
    ax = fig.gca()

    # Reference (perfect) mapping
    perfect = np.linspace(np.min(Y), np.max(Y), 20)
    ax.plot(perfect, np.zeros_like(perfect), linestyle='--', color='b', label="Perfect mapping")

    # For this param
    offset = mu - Y
    ax.scatter(Y, offset, color='tab:blue', marker='o')
    ax.set_title('offset vs. truth ({:s})'.format(name))
    ax.set_ylabel('pred - truth')
    ax.set_xlabel('truth')
    return fig

def _sigmoid(self, x):
    return 1.0/(np.exp(-x) + 1.0)

def get_mae(pred_mu, true_mu, Y_cols):
    """Get the total RMSE of predicted mu of the primary Gaussian wrt the transformed labels mu in a batch of validation data

    Parameters
    ----------
    pred_mu : np.array of shape `[batch_size, Y_dim]`
        predicted means of the primary Gaussian
    true_mu : np.array of shape `[batch_size, Y_dim]`
        true (label) Gaussian means
    Y_cols : np.array of shape `[Y_dim,]`
        the column names

    Returns
    -------
    dict
        total mean of the RMSE for that batch

    """
    abs_error = np.abs(pred_mu - true_mu) # [batch_size, Y_dim]
    mae_dict = {}
    # Total sq error, averaged across examples
    mae_dict['mae'] = np.median(np.sum(abs_error, axis=1)) # float
    # Parameter-wise sq error, averaged across examples
    mae_param = np.median(abs_error, axis=0) # [Y_dim,]
    param_dict = dict(zip(Y_cols, range(len(Y_cols))))
    for k, i in param_dict.items():
        mae_dict[k] = mae_param[i]
    return mae_dict

def interpret_pred(pred, Y_dim):
    """Slice the network prediction into means and cov matrix elements

    Parameters
    ----------
    pred : np.array of shape `[batch_size, out_dim]`
    Y_dim : int
        number of parameters to predict

    Note
    ----
    Currently hardcoded for `DoubleGaussianNLL`. (Update: no longer used; slicing function replaced by the BNNPosterior class.)

    Returns
    -------
    dict
        pred sliced into parameters of the Gaussians to predict

    """
    d = Y_dim # for readability
    mu = pred[:, :d]
    logvar = pred[:, d:2*d]
    F = pred[:, 2*d:4*d]
    mu2 = pred[:, 4*d:5*d]
    logvar2 = pred[:, 5*d:6*d]
    F2 = pred[:, 6*d:8*d]
    alpha = pred[:, -1]
    w2 = 0.5/(np.exp(-alpha) + 1.0)
    pred_dict = dict(
                     mu=mu,
                     logvar=logvar,
                     F=F,
                     mu2=mu2,
                     logvar2=logvar2,
                     F2=F2,
                     w2=w2,
                     )
    return pred_dict




import os
import sys
import argparse
import random
from addict import Dict
import numpy as np
import torch

__all__ = ['parse_inference_args', 'seed_everything', 'HiddenPrints']

def parse_inference_args():
    """Parse command-line arguments
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('test_config_file_path', help='path to the user-defined test config file')
    parser.add_argument('--lens_indices_path', default=None, dest='lens_indices_path', type=str,
                        help='path to a text file with specific lens indices to test on (Default: None)')
    args = parser.parse_args()
    # sys.argv rerouting for setuptools entry point
    if args is None:
        args = Dict()
        args.user_cfg_path = sys.argv[0]
        #args.n_data = sys.argv[1]
    return args

def seed_everything(global_seed):
    """Seed everything for reproducibility

    global_seed : int
        seed for `np.random`, `random`, and relevant `torch` backends

    """
    np.random.seed(global_seed)
    random.seed(global_seed)
    torch.manual_seed(global_seed)
    torch.cuda.manual_seed(global_seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False

class HiddenPrints:
    """Hide standard output

    """
    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout



"""Script to run MCMC cosmological sampling for individual lenses, using the BNN posterior

It borrows heavily from the `catalogue modelling.ipynb` notebook in Lenstronomy Extensions, which you can find `here <https://github.com/sibirrer/lenstronomy_extensions/blob/master/lenstronomy_extensions/Notebooks/catalogue%20modelling.ipynb>`_.

Example
-------
To run this script, pass in the path to the user-defined inference config file as the argument::
    
    $ python h0rton/infer_h0_mcmc_default.py mcmc_default.json

"""

import numpy as np
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver
from lenstronomy.Workflow.fitting_sequence import FittingSequence
import lenstronomy.Util.util as util
from baobab.sim_utils.psf_utils import get_PSF_model
from baobab.data_augmentation import noise_lenstronomy
__all__ = ["ForwardModelingPosterior"]


class ForwardModelingPosterior():
    """Joint posterior over the model parameters of lens mass, lens light, source light, and AGN point source
    as well as D_dt (H0)

    """
    def __init__(self, kwargs_lens_eqn_solver=None, astrometric_sigma=0.005, supersampling_factor=1):
        """Configure reasonable MCMC parameter kwargs and numerics common to all lenses

        """
        # Define assumed model profiles, with MCMC kwargs
        self.kwargs_model = dict(lens_model_list=['PEMD', 'SHEAR'],
                                point_source_model_list=['LENSED_POSITION'],
                                lens_light_model_list=['SERSIC_ELLIPSE'],
                                source_light_model_list=['SERSIC_ELLIPSE'])
        lens_params = self.get_lens_mass_mcmc_kwargs()
        lens_light_params = self.get_lens_light_mcmc_kwargs()
        source_params = self.get_source_mcmc_kwargs()
        ps_params = self.get_ps_mcmc_kwargs(lens_params[0])
        cosmo_params = self.get_cosmo_mcmc_kwargs()
        self.kwargs_params = {'lens_model': lens_params,
                            'source_model': source_params,
                            'lens_light_model': lens_light_params,
                            'point_source_model': ps_params,
                            'special': cosmo_params}
        # Define MCMC numerics
        self.kwargs_numerics = {'supersampling_factor': supersampling_factor,
                                'supersampling_convolution': False}
        self.num_source_model = len(self.kwargs_model['source_light_model_list'])
        self.kwargs_constraints = dict(
                                       joint_source_with_point_source=[[0, 0]], # AGN and host share centers
                                       joint_lens_with_light=[[0, 0, ['center_x', 'center_y']]], # Lens mass and light share centers (some nonlinear solvers override this assumption)
                                       #joint_lens_with_lens=[[0, 1, ['center_x', 'ra_0'], [0, 1, ['center_y', 'dec_0']]]],
                                       Ddt_sampling=True,
                                       )
        if kwargs_lens_eqn_solver is None:
            kwargs_lens_eqn_solver = {'min_distance': 0.05, 'search_window': 5.12, 'num_iter_max': 200}
        n_img = len(ps_params[0][0]['ra_image'])
        self.kwargs_likelihood = {'check_bounds': True,
                                  'sort_images_by_dec': True,
                                  'prior_lens': [],
                                  'prior_special': [],
                                 'force_no_add_image': False,
                                 'custom_logL_addition': self.couple_lens_mass_light_centroids if n_img == 4 else None, # necessary for quads for which the PROFILE_SHEAR solver ignores the joint_lens_with_light constraint
                                 'source_marg': False,
                                 'image_position_uncertainty': astrometric_sigma,
                                 'check_matched_source_position': True,
                                 'source_position_tolerance': 0.001,
                                 'time_delay_likelihood': True,
                                 'kwargs_lens_eqn_solver': kwargs_lens_eqn_solver,
                                         }

    def couple_lens_mass_light_centroids(self, kwargs_lens, kwargs_source, kwargs_lens_light=None, kwargs_ps=None, kwargs_special=None, kwargs_extinction=None):
        """Force the lens mass and lens light to share centroids by appending to the MCMC objective

        """
        return -(kwargs_lens[0]['center_x'] - kwargs_lens_light[0]['center_x'])**2.0 - (kwargs_lens[0]['center_y'] - kwargs_lens_light[0]['center_y'])**2.0

    def set_kwargs_data_joint(self, image, measured_td, measured_td_sigma, survey_object_dict, eff_exposure_time=5400.0, inverse=False):
        """Feed in time delay and imaging data, different across lenses

        Parameters
        ----------
        image : np.array, of shape [n_pix, n_pix]
        measured_td : np.array, of shape [n_img - 1,]
        measured_td_sigma : float or np.array of shape [n_img - 1,]

        """
        num_pix = image.shape[0]
        for i, (bp, survey_object) in enumerate(survey_object_dict.items()): # FIXME: only single band
            noise_kwargs = survey_object.kwargs_single_band()
            noise_kwargs['exposure_time'] = eff_exposure_time
            psf_kernel_size = survey_object.psf_kernel_size
            which_psf_maps = survey_object.which_psf_maps
        _, _, ra_0, dec_0, _, _, Mpix2coord, _ = util.make_grid_with_coordtransform(numPix=num_pix,
                                                                                    deltapix=noise_kwargs['pixel_scale'],
                                                                                    center_ra=0,
                                                                                    center_dec=0,
                                                                                    subgrid_res=1,
                                                                                    inverse=inverse)
        noise_dict = noise_lenstronomy.get_noise_sigma2_lenstronomy(image, **noise_kwargs)
        pixel_rms = (noise_dict['sky'] + noise_dict['readout'])**0.5
        total_exptime = noise_kwargs['exposure_time']*noise_kwargs['num_exposures']
        kwargs_data = dict(
                           background_rms=pixel_rms,
                           exposure_time=total_exptime,
                           ra_at_xy_0=ra_0,
                           dec_at_xy_0=dec_0,
                           transform_pix2angle=Mpix2coord,
                           image_data=image,
                           )
        psf_model = get_PSF_model(psf_type=noise_kwargs['psf_type'], pixel_scale=noise_kwargs['pixel_scale'], seeing=noise_kwargs['seeing'], kernel_size=psf_kernel_size, which_psf_maps=which_psf_maps)
        kwargs_psf = dict(
                          psf_type=psf_model.psf_type,
                          pixel_size=psf_model._pixel_size,
                          kernel_point_source=psf_model._kernel_point_source,
                          )
        image_band = [kwargs_data, kwargs_psf, self.kwargs_numerics]
        self.multi_band_list = [image_band] # single band
        self.kwargs_data_joint = dict(
                                      multi_band_list=self.multi_band_list,
                                      multi_band_type='multi-linear',
                                      time_delays_measured=measured_td,
                                      time_delays_uncertainties=measured_td_sigma
                                      )

    def get_lens_mass_mcmc_kwargs(self):
        """Compile reasonable MCMC kwargs for the lens mass components, PEMD and SHEAR

        """
        fixed_lens = []
        kwargs_lens_init = []
        kwargs_lens_sigma = []
        kwargs_lower_lens = []
        kwargs_upper_lens = []
        # Populate PEMD kwargs
        fixed_lens.append({}) 
        kwargs_lens_init.append({'theta_E': 1.0, 'e1': 0.0, 'e2': 0.0, 'gamma': 2.0, 'center_x': 0.0, 'center_y': 0.0})
        kwargs_lens_sigma.append({'theta_E': 0.2, 'e1': 0.1, 'e2': 0.1, 'gamma': 0.1, 'center_x': 0.01, 'center_y': 0.01})
        kwargs_lower_lens.append({'theta_E': 0.01, 'e1': -0.5, 'e2': -0.5, 'gamma': 1.5, 'center_x': -10, 'center_y': -10})
        kwargs_upper_lens.append({'theta_E': 10, 'e1': 0.5, 'e2': 0.5, 'gamma': 2.5, 'center_x': 10, 'center_y': 10})
        # Populate SHEAR kwargs
        fixed_lens.append({'ra_0': 0, 'dec_0': 0}) #fixed_lens.append({}) 
        kwargs_lens_init.append({'gamma1': 0.0, 'gamma2': 0.0}) #'ra_0': 0.0, 'dec_0': 0.0})
        kwargs_lens_sigma.append({'gamma1': 0.1, 'gamma2': 0.1}) #'ra_0': 0.05, 'dec_0': 0.05})
        kwargs_lower_lens.append({'gamma1': -0.5, 'gamma2': -0.5}) #'ra_0': -10.0, 'dec_0': -10.0})
        kwargs_upper_lens.append({'gamma1': 0.5, 'gamma2': 0.5}) #'ra_0': 10.0, 'dec_0': 10.0})
        # Collect into list
        lens_params = [kwargs_lens_init, kwargs_lens_sigma, fixed_lens, kwargs_lower_lens, kwargs_upper_lens]
        return lens_params

    def get_lens_light_mcmc_kwargs(self):
        """Compile reasonable MCMC kwargs for the SERSIC_ELLIPSE lens light

        """
        fixed_lens_light = []
        kwargs_lens_light_init = []
        kwargs_lens_light_sigma = []
        kwargs_lower_lens_light = []
        kwargs_upper_lens_light = []
        # Populate SERSIC_ELLIPSE kwargs
        fixed_lens_light.append({})
        kwargs_lens_light_init.append({'R_sersic': 0.5, 'n_sersic': 3.0, 'e1': 0.0, 'e2': 0.0, 'center_x': 0.0, 'center_y': 0.0})
        kwargs_lens_light_sigma.append({'R_sersic': 0.1, 'n_sersic': 0.5, 'e1': 0.1, 'e2': 0.1, 'center_x': 0.1, 'center_y': 0.1})
        kwargs_lower_lens_light.append({'e1': -0.5, 'e2': -0.5, 'R_sersic': 0.01, 'n_sersic': 0.5, 'center_x': -10, 'center_y': -10})
        kwargs_upper_lens_light.append({'e1': 0.5, 'e2': 0.5, 'R_sersic': 10, 'n_sersic': 8, 'center_x': 10, 'center_y': 10})
        # Collect into list
        lens_light_params = [kwargs_lens_light_init, kwargs_lens_light_sigma, fixed_lens_light, kwargs_lower_lens_light, kwargs_upper_lens_light]
        return lens_light_params

    def get_source_mcmc_kwargs(self):
        """Compile reasonable MCMC kwargs for the SERSIC_ELLIPSE source light

        """
        fixed_source = []
        kwargs_source_init = []
        kwargs_source_sigma = []
        kwargs_lower_source = []
        kwargs_upper_source = []
        # Populate SERSIC_ELLIPSE kwargs
        fixed_source.append({})
        kwargs_source_init.append({'R_sersic': 0.3, 'n_sersic': 2.0, 'e1': 0.0, 'e2': 0.0, 'center_x': 0.0, 'center_y': 0.0})
        kwargs_source_sigma.append({'R_sersic': 0.1, 'n_sersic': 0.5, 'e1': 0.1, 'e2': 0.1, 'center_x': 0.1, 'center_y': 0.1})
        kwargs_lower_source.append({'e1': -0.5, 'e2': -0.5, 'R_sersic': 0.001, 'n_sersic': .1, 'center_x': -10, 'center_y': -10})
        kwargs_upper_source.append({'e1': 0.5, 'e2': 0.5, 'R_sersic': 1.0, 'n_sersic': 5., 'center_x': 10, 'center_y': 10})
        # Collect into list
        source_params = [kwargs_source_init, kwargs_source_sigma, fixed_source, kwargs_lower_source, kwargs_upper_source]
        return source_params

    def get_ps_mcmc_kwargs(self, lens_init_kwargs):
        """Get the init kwargs for point source LENSED_POSITION based on the init kwargs for the lens model,
        a centered point source, and reasonable redshifts 

        Returns
        -------
        dict
            kwargs for LENSED_POSITION that can be used to initialize the MCMC
        
        """
        lens_model_class = LensModel(lens_model_list=['PEMD', 'SHEAR'], z_lens=0.5, z_source=1.5)
        kwargs_lens = lens_init_kwargs
        lensEquationSolver = LensEquationSolver(lens_model_class)
        x_image, y_image = lensEquationSolver.findBrightImage(0.0, 0.0, # center the point source
                                                              kwargs_lens, numImages=4,
                                                              min_distance=0.05, search_window=5.12)
        magnification = lens_model_class.magnification(x_image, y_image, kwargs=kwargs_lens)
        fixed_ps = [{}]
        kwargs_ps_init = [{'ra_image': x_image, 'dec_image': y_image, 'point_amp': np.abs(magnification)*1000}]  
        kwargs_ps_sigma = [{'ra_image': 0.01 * np.ones(len(x_image)), 'dec_image': 0.01 * np.ones(len(x_image))}]
        kwargs_lower_ps = [{'ra_image': -10 * np.ones(len(x_image)), 'dec_image': -10 * np.ones(len(y_image))}]
        kwargs_upper_ps = [{'ra_image': 10* np.ones(len(x_image)), 'dec_image': 10 * np.ones(len(y_image))}]
        ps_params = [kwargs_ps_init, kwargs_ps_sigma, fixed_ps, kwargs_lower_ps, kwargs_upper_ps]
        return ps_params

    def get_cosmo_mcmc_kwargs(self):
        """Compile reasonable MCMC kwargs for D_dt

        """
        fixed_cosmo = {}
        kwargs_cosmo_init = {'D_dt': 5000}
        kwargs_cosmo_sigma = {'D_dt': 3000}
        kwargs_lower_cosmo = {'D_dt': 0}
        kwargs_upper_cosmo = {'D_dt': 15000}
        cosmo_params = [kwargs_cosmo_init, kwargs_cosmo_sigma, fixed_cosmo, kwargs_lower_cosmo, kwargs_upper_cosmo]
        return cosmo_params

    def run_mcmc(self, mcmc_numerics):
        """Sample from the joint likelihood

        """
        fitting_seq = FittingSequence(self.kwargs_data_joint, 
                                      self.kwargs_model, 
                                      self.kwargs_constraints, 
                                      self.kwargs_likelihood, 
                                      self.kwargs_params)
        fitting_kwargs_list = [['MCMC', mcmc_numerics]]
        #with script_utils.HiddenPrints():
        chain_list_mcmc = fitting_seq.fit_sequence(fitting_kwargs_list)
        kwargs_result_mcmc = fitting_seq.best_fit()
        return chain_list_mcmc, kwargs_result_mcmc

    @property
    def kwargs_lens_init(self):
        return self.kwargs_params['lens_model'][0][0]

    @kwargs_lens_init.setter
    def kwargs_lens_init(self, new_kwargs_lens_init):
        self.kwargs_params['lens_model'][0][0] = new_kwargs_lens_init[0]
        self.kwargs_params['lens_model'][0][1] = new_kwargs_lens_init[1]

    @property
    def kwargs_lens_light_init(self):
        return self.kwargs_params['lens_light_model'][0][0]

    @kwargs_lens_light_init.setter
    def kwargs_lens_light_init(self, new_kwargs_lens_light_init):
        self.kwargs_params['lens_light_model'][0][0] = new_kwargs_lens_light_init

    @property
    def kwargs_source_init(self):
        return self.kwargs_params['source_model'][0][0]

    @kwargs_source_init.setter
    def kwargs_source_init(self, new_kwargs_source_init):
        self.kwargs_params['source_model'][0][0] = new_kwargs_source_init

    @property
    def kwargs_ps_init(self):
        return self.kwargs_params['point_source_model'][0][0]

    @kwargs_ps_init.setter
    def kwargs_ps_init(self, new_kwargs_ps_init):
        self.kwargs_params['point_source_model'][0][0] = new_kwargs_ps_init

    @property
    def kwargs_special_init(self):
        return self.kwargs_params['special'][0][0]

    @kwargs_special_init.setter
    def kwargs_special_init(self, new_kwargs_special_init):
        self.kwargs_params['special'][0][0] = new_kwargs_special_init


# -*- coding: utf-8 -*-
"""Training the Bayesian neural network (BNN).
This script performs H0 inference on a test (or validation) sample using simple MC sampling

Example
-------
To run this script, pass in the path to the user-defined training config file as the argument::
    
    $ infer_h0 h0rton/h0_inference_config.json

"""
import os
import time
from tqdm import tqdm
from ast import literal_eval
import numpy as np
import pandas as pd
import scipy.stats as stats
import torch
from torch.utils.data import DataLoader
import h0rton.models
# Baobab modules
from baobab import BaobabConfig
# H0rton modules
from h0rton.configs import TrainValConfig, TestConfig
import h0rton.losses
import h0rton.script_utils as script_utils
from h0rton.h0_inference import H0Posterior, plot_weighted_h0_histogram
from h0rton.trainval_data import XYData
from astropy.cosmology import FlatLambdaCDM

def main():
    args = script_utils.parse_inference_args()
    test_cfg = TestConfig.from_file(args.test_config_file_path)
    baobab_cfg = BaobabConfig.from_file(test_cfg.data.test_baobab_cfg_path)
    cfg = TrainValConfig.from_file(test_cfg.train_val_config_file_path)
    # Set device and default data type
    device = torch.device(test_cfg.device_type)
    if device.type == 'cuda':
        torch.set_default_tensor_type('torch.cuda.FloatTensor')
    else:
        torch.set_default_tensor_type('torch.FloatTensor')
    script_utils.seed_everything(test_cfg.global_seed)
    
    ############
    # Data I/O #
    ############
    train_data = XYData(is_train=True, 
                        Y_cols=cfg.data.Y_cols, 
                        float_type=cfg.data.float_type, 
                        define_src_pos_wrt_lens=cfg.data.define_src_pos_wrt_lens, 
                        rescale_pixels=cfg.data.rescale_pixels, 
                        log_pixels=cfg.data.log_pixels, 
                        add_pixel_noise=cfg.data.add_pixel_noise, 
                        eff_exposure_time=cfg.data.eff_exposure_time, 
                        train_Y_mean=None, 
                        train_Y_std=None, 
                        train_baobab_cfg_path=cfg.data.train_baobab_cfg_path, 
                        val_baobab_cfg_path=test_cfg.data.test_baobab_cfg_path, 
                        for_cosmology=False)
    # Define val data and loader
    test_data = XYData(is_train=False, 
                       Y_cols=cfg.data.Y_cols, 
                       float_type=cfg.data.float_type, 
                       define_src_pos_wrt_lens=cfg.data.define_src_pos_wrt_lens, 
                       rescale_pixels=cfg.data.rescale_pixels, 
                       log_pixels=cfg.data.log_pixels, 
                       add_pixel_noise=cfg.data.add_pixel_noise, 
                       eff_exposure_time=cfg.data.eff_exposure_time, 
                       train_Y_mean=train_data.train_Y_mean, 
                       train_Y_std=train_data.train_Y_std, 
                       train_baobab_cfg_path=cfg.data.train_baobab_cfg_path, 
                       val_baobab_cfg_path=test_cfg.data.test_baobab_cfg_path, 
                       for_cosmology=True)
    cosmo_df = test_data.Y_df
    if test_cfg.data.lens_indices is None:
        n_test = test_cfg.data.n_test # number of lenses in the test set
        lens_range = range(n_test)
    else: # if specific lenses are specified
        lens_range = test_cfg.data.lens_indices
        n_test = len(lens_range)
        print("Performing H0 inference on {:d} specified lenses...".format(n_test))
    batch_size = max(lens_range) + 1
    test_loader = DataLoader(test_data, batch_size=batch_size, shuffle=False, drop_last=True)
    # Output directory into which the H0 histograms and H0 samples will be saved
    out_dir = test_cfg.out_dir
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
        print("Destination folder path: {:s}".format(out_dir))
    else:
        raise OSError("Destination folder already exists.")

    ######################
    # Load trained state #
    ######################
    # Instantiate loss function
    loss_fn = getattr(h0rton.losses, cfg.model.likelihood_class)(Y_dim=train_data.Y_dim, device=device)
    # Instantiate posterior (for logging)
    bnn_post = getattr(h0rton.h0_inference.gaussian_bnn_posterior, loss_fn.posterior_name)(train_data.Y_dim, device, train_data.train_Y_mean, train_data.train_Y_std)
    with torch.no_grad(): # TODO: skip this if lens_posterior_type == 'truth'
        for X_, Y_ in test_loader:
            X = X_.to(device)
            Y = Y_.to(device)
            break

    # Export the input images X for later error analysis
    if test_cfg.export.images:
        for lens_i in range(n_test):
            X_img_path = os.path.join(out_dir, 'X_{0:04d}.npy'.format(lens_i))
            np.save(X_img_path, X[lens_i, 0, :, :].cpu().numpy())

    ################
    # H0 Posterior #
    ################
    h0_prior = getattr(stats, test_cfg.h0_prior.dist)(**test_cfg.h0_prior.kwargs)
    kappa_ext_prior = getattr(stats, test_cfg.kappa_ext_prior.dist)(**test_cfg.kappa_ext_prior.kwargs)
    aniso_param_prior = getattr(stats, test_cfg.aniso_param_prior.dist)(**test_cfg.aniso_param_prior.kwargs)
    # FIXME: hardcoded
    kwargs_model = dict(
                        lens_model_list=['PEMD', 'SHEAR'],
                        lens_light_model_list=['SERSIC_ELLIPSE'],
                        source_light_model_list=['SERSIC_ELLIPSE'],
                        point_source_model_list=['SOURCE_POSITION'],
                        cosmo=FlatLambdaCDM(H0=70.0, Om0=0.3)
                       #'point_source_model_list' : ['LENSED_POSITION']
                       )
    h0_post = H0Posterior(
                          H0_prior=h0_prior,
                          kappa_ext_prior=kappa_ext_prior,
                          aniso_param_prior=aniso_param_prior,
                          exclude_vel_disp=test_cfg.h0_posterior.exclude_velocity_dispersion,
                          kwargs_model=kwargs_model,
                          baobab_time_delays=test_cfg.time_delay_likelihood.baobab_time_delays,
                          kinematics=baobab_cfg.bnn_omega.kinematics,
                          kappa_transformed=test_cfg.kappa_ext_prior.transformed,
                          Om0=baobab_cfg.bnn_omega.cosmology.Om0,
                          define_src_pos_wrt_lens=cfg.data.define_src_pos_wrt_lens,
                          kwargs_lens_eqn_solver={'min_distance': 0.05, 'search_window': baobab_cfg.instrument['pixel_scale']*baobab_cfg.image['num_pix'], 'num_iter_max': 100}
                          )
    # Get H0 samples for each system
    if not test_cfg.time_delay_likelihood.baobab_time_delays:
        if 'abcd_ordering_i' not in cosmo_df:
            raise ValueError("If the time delay measurements were not generated using Baobab, the user must specify the order of image positions in which the time delays are listed, in order of increasing dec.")
    required_params = h0_post.required_params

    ########################
    # Lens Model Posterior #
    ########################
    n_samples = test_cfg.h0_posterior.n_samples # number of h0 samples per lens
    sampling_buffer = test_cfg.h0_posterior.sampling_buffer # FIXME: dynamically sample more if we run out of samples
    actual_n_samples = int(n_samples*sampling_buffer)

    # Add artificial noise around the truth values
    Y_orig = bnn_post.transform_back_mu(Y).cpu().numpy().reshape(batch_size, test_data.Y_dim)
    Y_orig_df = pd.DataFrame(Y_orig, columns=cfg.data.Y_cols)
    Y_orig_values = Y_orig_df[required_params].values[:, np.newaxis, :] # [n_test, 1, Y_dim]
    artificial_noise = np.random.randn(batch_size, actual_n_samples, test_data.Y_dim)*Y_orig_values*test_cfg.fractional_error_added_to_truth # [n_test, buffer*n_samples, Y_dim]
    lens_model_samples_values = Y_orig_values + artificial_noise # [n_test, buffer*n_samples, Y_dim]

    # Placeholders for mean and std of H0 samples per system
    mean_h0_set = np.zeros(n_test)
    std_h0_set = np.zeros(n_test)
    inference_time_set = np.zeros(n_test)
    # For each lens system...
    total_progress = tqdm(total=n_test)
    sampling_progress = tqdm(total=n_samples)
    prerealized_time_delays = test_cfg.error_model.prerealized_time_delays
    if prerealized_time_delays:
        realized_time_delays = pd.read_csv(test_cfg.error_model.realized_time_delays_path, index_col=None)
    else:
        realized_time_delays = pd.DataFrame()
        realized_time_delays['measured_td_wrt0'] = [[]]*len(lens_range)
    for i, lens_i in enumerate(lens_range):
        lens_i_start_time = time.time()
        # Each lens gets a unique random state for td and vd measurement error realizations.
        rs_lens = np.random.RandomState(lens_i)
        # BNN samples for lens_i
        bnn_sample_df = pd.DataFrame(lens_model_samples_values[lens_i, :, :], columns=required_params)
        # Cosmology observables for lens_i
        cosmo = cosmo_df.iloc[lens_i]
        true_td = np.array(literal_eval(cosmo['true_td']))
        true_img_dec = np.array(literal_eval(cosmo['y_image']))
        true_img_ra = np.array(literal_eval(cosmo['x_image']))
        increasing_dec_i = np.argsort(true_img_dec)
        true_img_dec = true_img_dec[increasing_dec_i]
        true_img_ra = true_img_ra[increasing_dec_i]        
        measured_vd = cosmo['true_vd']*(1.0 + rs_lens.randn()*test_cfg.error_model.velocity_dispersion_frac_error)
        if prerealized_time_delays:
            measured_td_wrt0 = np.array(literal_eval(realized_time_delays.iloc[lens_i]['measured_td_wrt0']))
        else:
            true_td = true_td[increasing_dec_i]
            true_td = true_td[1:] - true_td[0]
            measured_td_wrt0 = true_td + rs_lens.randn(*true_td.shape)*test_cfg.error_model.time_delay_error # [n_img -1,]
            realized_time_delays.at[lens_i, 'measured_td_wrt0'] = list(measured_td_wrt0)
        #print("True: ", true_td)
        #print("True img: ", true_img_dec)
        #print("measured td: ", measured_td_wrt0)
        h0_post.set_cosmology_observables(
                                          z_lens=cosmo['z_lens'], 
                                          z_src=cosmo['z_src'], 
                                          measured_vd=measured_vd, 
                                          measured_vd_err=test_cfg.velocity_dispersion_likelihood.sigma, 
                                          measured_td_wrt0=measured_td_wrt0,
                                          measured_td_err=test_cfg.time_delay_likelihood.sigma, 
                                          abcd_ordering_i=np.arange(len(true_td) + 1),
                                          true_img_dec=true_img_dec,
                                          true_img_ra=true_img_ra,
                                          kappa_ext=cosmo['kappa_ext'], # not necessary
                                          )
        h0_post.set_truth_lens_model(sampled_lens_model_raw=bnn_sample_df.iloc[0])
        # Initialize output array
        h0_samples = np.full(n_samples, np.nan)
        h0_weights = np.zeros(n_samples)
        # For each sample from the lens model posterior of this lens system...
        sampling_progress.reset()
        valid_sample_i = 0
        sample_i = 0
        while valid_sample_i < n_samples:
            if sample_i > actual_n_samples - 1:
                break
            #try:
            # Each sample for a given lens gets a unique random state for H0, k_ext, and aniso_param realizations.
            rs_sample = np.random.RandomState(int(str(lens_i) + str(sample_i).zfill(5)))
            h0, weight = h0_post.get_h0_sample_truth(rs_sample)
            h0_samples[valid_sample_i] = h0
            h0_weights[valid_sample_i] = weight
            sampling_progress.update(1)
            time.sleep(0.001)
            valid_sample_i += 1
            sample_i += 1
            #except:
            #    sample_i += 1
            #    continue
        sampling_progress.refresh()
        lens_i_end_time = time.time()
        inference_time = (lens_i_end_time - lens_i_start_time)/60.0 # min
        h0_dict = dict(
                       h0_samples=h0_samples,
                       h0_weights=h0_weights,
                       n_sampling_attempts=sample_i,
                       measured_td_wrt0=measured_td_wrt0,
                       inference_time=inference_time
                       )
        h0_dict_save_path = os.path.join(out_dir, 'h0_dict_{0:04d}.npy'.format(lens_i))
        np.save(h0_dict_save_path, h0_dict)
        h0_stats = plot_weighted_h0_histogram(h0_samples, h0_weights, lens_i, cosmo['H0'], include_fit_gaussian=test_cfg.plotting.include_fit_gaussian, save_dir=out_dir)
        mean_h0_set[i] = h0_stats['mean']
        std_h0_set[i] = h0_stats['std']
        inference_time_set[i] = inference_time
        total_progress.update(1)
    total_progress.close()
    if not prerealized_time_delays:
        realized_time_delays.to_csv(os.path.join(out_dir, 'realized_time_delays.csv'), index=None)
    h0_stats = dict(
                    mean=mean_h0_set,
                    std=std_h0_set,
                    inference_time=inference_time_set,
                    )
    h0_stats_save_path = os.path.join(out_dir, 'h0_stats')
    np.save(h0_stats_save_path, h0_stats)

if __name__ == '__main__':
    main()


from setuptools import setup, find_packages
#print(find_packages())
#required_packages = []
#with open('requirements.txt') as f:
#    required_packages = f.read().splitlines()
#required_packages += ['corner.py @ https://github.com/jiwoncpark/corner.py/archive/master.zip']
#print(required_packages)

setup(
      name='h0rton',
      version='v1.0',
      author='Ji Won Park',
      author_email='jiwon.christine.park@gmail.com',
      packages=find_packages(),
      license='LICENSE.md',
      description='Bayesian neural network for hierarchical inference of the Hubble constant',
      long_description=open("README.rst").read(),
      long_description_content_type='text/x-rst',
      url='https://github.com/jiwoncpark/h0rton',
      #install_requires=required_packages,
      #dependency_links=['http://github.com/jiwoncpark/corner.py/tarball/master#egg=corner_jiwoncpark'],
      include_package_data=True,
      entry_points={
      'console_scripts': ['train=h0rton.train:main', 'infer_h0=h0rton.infer_h0_mcmc_default:main'],
      },
      test_suite='nose.collector',
      tests_require=['nose'],
      classifiers=['Development Status :: 4 - Beta',
      'License :: OSI Approved :: BSD License',
      'Intended Audience :: Developers',
      'Intended Audience :: Science/Research',
      'Operating System :: OS Independent',
      'Programming Language :: Python'],
      keywords='physics'
      )
